@@ -1,14 +1,16 @@
+from __future__ import annotations
 import logging
-from typing import Any, Callable, Optional, Set
+from typing import Any
+from collections.abc import Callable
 
 import cle
 from cle.backends.externs import KernelObject, ExternObject
 from cle.backends.tls.elf_tls import ELFTLSObject
-
 from sortedcontainers import SortedDict
 
-from ...knowledge_plugins.cfg.memory_data import MemoryDataSort, MemoryData
-from ..analysis import Analysis
+from angr.analyses import AnalysesHub
+from angr.knowledge_plugins.cfg.memory_data import MemoryDataSort, MemoryData
+from angr.analyses.analysis import Analysis
 
 _l = logging.getLogger(name=__name__)
 
@@ -34,7 +36,7 @@ class CFBlanketView:
                 addr_ += obj
                 # Find gaps
                 # TODO: finish it
-                raise NotImplementedError()
+                raise NotImplementedError
 
 
 #
@@ -77,8 +79,7 @@ class Unknown:
             raise Exception("You cannot create an unknown region of size 0.")
 
     def __repr__(self):
-        s = f"<Unknown {self.addr:#x}-{self.addr + self.size:#x}>"
-        return s
+        return f"<Unknown {self.addr:#x}-{self.addr + self.size:#x}>"
 
 
 class CFBlanket(Analysis):
@@ -95,8 +96,8 @@ class CFBlanket(Analysis):
 
     def __init__(
         self,
-        exclude_region_types: Optional[Set[str]] = None,
-        on_object_added: Optional[Callable[[int, Any], None]] = None,
+        exclude_region_types: set[str] | None = None,
+        on_object_added: Callable[[int, Any], None] | None = None,
     ):
         """
         :param on_object_added: Callable with parameters (addr, obj) called after an object is added to the blanket.
@@ -105,7 +106,7 @@ class CFBlanket(Analysis):
 
         self._on_object_added_callback = on_object_added
         self._regions = []
-        self._exclude_region_types = set() if not exclude_region_types else exclude_region_types
+        self._exclude_region_types = exclude_region_types if exclude_region_types else set()
 
         self._init_regions()
 
@@ -118,19 +119,19 @@ class CFBlanket(Analysis):
     def _init_regions(self):
         for obj in self.project.loader.all_objects:
             if isinstance(obj, cle.MetaELF):
-                if obj.sections:
+                if obj.segments:
+                    if "segment" not in self._exclude_region_types:
+                        for segment in obj.segments:
+                            if segment.memsize > 0:
+                                mr = MemoryRegion(segment.vaddr, segment.memsize, "segment", obj, segment)
+                                self._regions.append(mr)
+                elif obj.sections:
                     if "section" not in self._exclude_region_types:
                         # Enumerate sections in an ELF file
                         for section in obj.sections:
                             if section.occupies_memory:
                                 mr = MemoryRegion(section.vaddr, section.memsize, "section", obj, section)
                                 self._regions.append(mr)
-                elif obj.segments:
-                    if "segment" not in self._exclude_region_types:
-                        for segment in obj.segments:
-                            if segment.memsize > 0:
-                                mr = MemoryRegion(segment.vaddr, segment.memsize, "segment", obj, segment)
-                                self._regions.append(mr)
                 else:
                     raise NotImplementedError(
                         "Currently ELFs without sections or segments are not supported. Please "
@@ -154,7 +155,7 @@ class CFBlanket(Analysis):
                     self._regions.append(mr)
             elif isinstance(obj, ExternObject):
                 if "extern" not in self._exclude_region_types:
-                    size = obj.max_addr - obj.min_addr
+                    size = obj.next_addr
                     mr = MemoryRegion(obj.min_addr, size, "extern", obj, None)
                     self._regions.append(mr)
             elif isinstance(obj, ELFTLSObject):
@@ -163,16 +164,13 @@ class CFBlanket(Analysis):
                     mr = MemoryRegion(obj.min_addr, size, "tls", obj, None)
                     self._regions.append(mr)
             else:
-                if hasattr(obj, "size"):
-                    size = obj.size
-                else:
-                    size = obj.max_addr - obj.min_addr
+                size = obj.size if hasattr(obj, "size") else obj.max_addr - obj.min_addr
                 type_ = "TODO"
                 mr = MemoryRegion(obj.min_addr, size, type_, obj, obj)
                 self._regions.append(mr)
 
         # Sort them just in case
-        self._regions = list(sorted(self._regions, key=lambda x: x.addr))
+        self._regions = sorted(self._regions, key=lambda x: x.addr)
 
     @property
     def regions(self):
@@ -185,8 +183,8 @@ class CFBlanket(Analysis):
     def floor_addr(self, addr):
         try:
             return next(self._blanket.irange(maximum=addr, reverse=True))
-        except StopIteration:
-            raise KeyError(addr)
+        except StopIteration as err:
+            raise KeyError(addr) from err
 
     def floor_item(self, addr):
         key = self.floor_addr(addr)
@@ -207,8 +205,8 @@ class CFBlanket(Analysis):
     def ceiling_addr(self, addr):
         try:
             return next(self._blanket.irange(minimum=addr))
-        except StopIteration:
-            raise KeyError(addr)
+        except StopIteration as err:
+            raise KeyError(addr) from err
 
     def ceiling_item(self, addr):
         key = self.ceiling_addr(addr)
@@ -259,14 +257,14 @@ class CFBlanket(Analysis):
                 if section.memsize == 0:
                     continue
                 min_addr, max_addr = section.min_addr, section.max_addr
-                output.append("### Object %s" % repr(section))
+                output.append(f"### Object {section!r}")
                 output.append(f"### Range {min_addr:#x}-{max_addr:#x}")
 
                 pos = min_addr
                 while pos < max_addr:
                     try:
                         addr, thing = self.floor_item(pos)
-                        output.append(f"{addr:#x}: {repr(thing)}")
+                        output.append(f"{addr:#x}: {thing!r}")
 
                         if thing.size == 0:
                             pos += 1
@@ -392,11 +390,8 @@ class CFBlanket(Analysis):
                 # impossible
                 raise Exception("Impossible")
 
-            if last_item.size == 0 or last_item.size is None:
-                # Make sure everything has a non-zero size
-                last_item_size = 1
-            else:
-                last_item_size = last_item.size
+            # Make sure everything has a non-zero size
+            last_item_size = 1 if last_item.size == 0 or last_item.size is None else last_item.size
             end_addr = last_addr + last_item_size
             if end_addr < max_addr:
                 try:
@@ -415,9 +410,9 @@ class CFBlanket(Analysis):
                                 obj,
                                 section,
                                 segment,
-                                next_addr,
+                                end_addr,
                             )
-                            bytes_ = self.project.loader.memory.load(next_addr, size)
+                            bytes_ = self.project.loader.memory.load(end_addr, size)
                         except KeyError:
                             # The address does not exist
                             bytes_ = None
@@ -429,7 +424,5 @@ class CFBlanket(Analysis):
                 addr = max_addr
 
 
-from angr.analyses import AnalysesHub
-
 AnalysesHub.register_default("CFB", CFBlanket)
 AnalysesHub.register_default("CFBlanket", CFBlanket)
