@@ -1,40 +1,31 @@
-from typing import List, Tuple, Iterable, Dict
+# pylint:disable=too-many-boolean-expressions
+from __future__ import annotations
+from collections.abc import Iterable
 import logging
 
 import archinfo
-import ailment
+import angr.ailment as ailment
 
-from ....calling_conventions import SimRegArg
-from ....code_location import CodeLocation
+from angr.calling_conventions import SimRegArg
+from angr.code_location import CodeLocation
+from angr.analyses.decompiler.stack_item import StackItem, StackItemType
 from .optimization_pass import OptimizationPass, OptimizationPassStage
 
 
 _l = logging.getLogger(name=__name__)
 
 
-def s2u(s, bits):
-    if s > 0:
-        return s
-    return (1 << bits) + s
-
-
 class RegisterSaveAreaSimplifier(OptimizationPass):
     """
     Optimizes away register spilling effects, including callee-saved registers.
+
+    This optimization runs between SSA-level0 and SSA-level1, which means registers are converted to vvars but stack
+    accesses stay unchanged.
     """
 
-    ARCHES = [
-        "X86",
-        "AMD64",
-        "ARM",
-        "ARMEL",
-        "ARMHF",
-        "ARMCortexM",
-        "MIPS32",
-        "MIPS64",
-    ]
-    PLATFORMS = ["cgc", "linux"]
-    STAGE = OptimizationPassStage.AFTER_GLOBAL_SIMPLIFICATION
+    ARCHES = None
+    PLATFORMS = None
+    STAGE = OptimizationPassStage.AFTER_SINGLE_BLOCK_SIMPLIFICATION
     NAME = "Simplify register save areas"
     DESCRIPTION = __doc__.strip()
 
@@ -59,29 +50,33 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
 
         return bool(info), {"info": info}
 
+    @staticmethod
+    def _modify_statement(
+        old_block, stmt_idx_: int, updated_blocks_, stack_offset: int | None = None
+    ):  # pylint:disable=unused-argument
+        if old_block not in updated_blocks_:
+            block = old_block.copy()
+            updated_blocks_[old_block] = block
+        else:
+            block = updated_blocks_[old_block]
+        block.statements[stmt_idx_] = None
+
     def _analyze(self, cache=None):
-        def _remove_statement(old_block, stmt_idx_: int, updated_blocks_):
-            if old_block not in updated_blocks_:
-                block = old_block.copy()
-                updated_blocks[old_block] = block
-            else:
-                block = updated_blocks[old_block]
-            block.statements[stmt_idx_] = None
 
         if cache is None:
             return
 
-        info: Dict[int, Dict[str, List[Tuple[int, CodeLocation]]]] = cache["info"]
+        info: dict[int, dict[str, list[tuple[int, CodeLocation]]]] = cache["info"]
         updated_blocks = {}
 
         for data in info.values():
             # remove storing statements
-            for _, codeloc in data["stored"]:
+            for stack_offset, codeloc in data["stored"]:
                 old_block = self._get_block(codeloc.block_addr, idx=codeloc.block_idx)
-                _remove_statement(old_block, codeloc.stmt_idx, updated_blocks)
-            for _, codeloc in data["restored"]:
+                self._modify_statement(old_block, codeloc.stmt_idx, updated_blocks, stack_offset=stack_offset)
+            for stack_offset, codeloc in data["restored"]:
                 old_block = self._get_block(codeloc.block_addr, idx=codeloc.block_idx)
-                _remove_statement(old_block, codeloc.stmt_idx, updated_blocks)
+                self._modify_statement(old_block, codeloc.stmt_idx, updated_blocks, stack_offset=stack_offset)
 
         for old_block, new_block in updated_blocks.items():
             # remove all statements that are None
@@ -89,7 +84,15 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
             # update it
             self._update_block(old_block, new_block)
 
-    def _find_registers_stored_on_stack(self) -> List[Tuple[int, int, CodeLocation]]:
+        if updated_blocks:
+            # update stack_items
+            for data in info.values():
+                for stack_offset, _ in data["stored"]:
+                    self.stack_items[stack_offset] = StackItem(
+                        stack_offset, self.project.arch.bytes, "regs", StackItemType.SAVED_REGS
+                    )
+
+    def _find_registers_stored_on_stack(self) -> list[tuple[int, int, CodeLocation]]:
         first_block = self._get_block(self._func.addr)
         if first_block is None:
             return []
@@ -100,17 +103,31 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
             if (
                 isinstance(stmt, ailment.Stmt.Store)
                 and isinstance(stmt.addr, ailment.Expr.StackBaseOffset)
-                and isinstance(stmt.data, ailment.Expr.Register)
+                and isinstance(stmt.addr.offset, int)
             ):
-                # it's storing registers to the stack!
-                stack_offset = stmt.addr.offset
-                reg_offset = stmt.data.reg_offset
-                codeloc = CodeLocation(first_block.addr, idx, block_idx=first_block.idx, ins_addr=stmt.ins_addr)
-                results.append((reg_offset, stack_offset, codeloc))
+                if isinstance(stmt.data, ailment.Expr.VirtualVariable) and stmt.data.was_reg:
+                    # it's storing registers to the stack!
+                    stack_offset = stmt.addr.offset
+                    reg_offset = stmt.data.reg_offset
+                    codeloc = CodeLocation(first_block.addr, idx, block_idx=first_block.idx, ins_addr=stmt.ins_addr)
+                    results.append((reg_offset, stack_offset, codeloc))
+                elif (
+                    self.project.arch.name == "AMD64"
+                    and isinstance(stmt.data, ailment.Expr.Convert)
+                    and isinstance(stmt.data.operand, ailment.Expr.VirtualVariable)
+                    and stmt.data.operand.was_reg
+                    and stmt.data.from_bits == 256
+                    and stmt.data.to_bits == 128
+                ):
+                    # storing xmm registers to the stack
+                    stack_offset = stmt.addr.offset
+                    reg_offset = stmt.data.operand.reg_offset
+                    codeloc = CodeLocation(first_block.addr, idx, block_idx=first_block.idx, ins_addr=stmt.ins_addr)
+                    results.append((reg_offset, stack_offset, codeloc))
 
         return results
 
-    def _find_registers_restored_from_stack(self) -> List[List[Tuple[int, int, CodeLocation]]]:
+    def _find_registers_restored_from_stack(self) -> list[list[tuple[int, int, CodeLocation]]]:
         all_results = []
         for ret_site in self._func.ret_sites + self._func.jumpout_sites:
             for block in self._get_blocks(ret_site.addr):
@@ -118,7 +135,8 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
                 for idx, stmt in enumerate(block.statements):
                     if (
                         isinstance(stmt, ailment.Stmt.Assignment)
-                        and isinstance(stmt.dst, ailment.Expr.Register)
+                        and isinstance(stmt.dst, ailment.Expr.VirtualVariable)
+                        and stmt.dst.was_reg
                         and isinstance(stmt.src, ailment.Expr.Load)
                         and isinstance(stmt.src.addr, ailment.Expr.StackBaseOffset)
                     ):
@@ -134,10 +152,10 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
 
     def _intersect_register_info(
         self,
-        stored: List[Tuple[int, int, CodeLocation]],
-        restored: Iterable[List[Tuple[int, int, CodeLocation]]],
-    ) -> Dict[int, Dict[str, List[Tuple[int, CodeLocation]]]]:
-        def _collect(info: List[Tuple[int, int, CodeLocation]], output, keystr: str):
+        stored: list[tuple[int, int, CodeLocation]],
+        restored: Iterable[list[tuple[int, int, CodeLocation]]],
+    ) -> dict[int, dict[str, list[tuple[int, CodeLocation]]]]:
+        def _collect(info: list[tuple[int, int, CodeLocation]], output, keystr: str):
             for reg_offset, stack_offset, codeloc in info:
                 if reg_offset not in output:
                     output[reg_offset] = {}
@@ -145,7 +163,7 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
                     output[reg_offset][keystr] = []
                 output[reg_offset][keystr].append((stack_offset, codeloc))
 
-        result: Dict[int, Dict[str, List[Tuple[int, CodeLocation]]]] = {}
+        result: dict[int, dict[str, list[tuple[int, CodeLocation]]]] = {}
         _collect(stored, result, "stored")
         for item in restored:
             _collect(item, result, "restored")
@@ -167,7 +185,7 @@ class RegisterSaveAreaSimplifier(OptimizationPass):
             lr_reg_offset = self.project.arch.registers["lr"][0]
         elif self.project.arch.name in {"MIPS32", "MIPS64"}:
             lr_reg_offset = self.project.arch.registers["ra"][0]
-        elif self.project.arch.name in {"PPC32", "PPC64"}:
+        elif self.project.arch.name in {"PPC32", "PPC64"} or self.project.arch.name.startswith("PowerPC:"):
             lr_reg_offset = self.project.arch.registers["lr"][0]
         else:
             lr_reg_offset = None
