@@ -1,26 +1,27 @@
 # pylint:disable=multiple-statements,line-too-long,consider-using-enumerate
-from typing import Dict, Set, Optional, Any, List, Union, Tuple, OrderedDict as ODict, TYPE_CHECKING
+from __future__ import annotations
+from typing import Any, TYPE_CHECKING
 import logging
 from collections import defaultdict, OrderedDict
 
 import networkx
 
 import claripy
-import ailment
+import angr.ailment as ailment
 
-from ....knowledge_plugins.cfg import IndirectJump, IndirectJumpType
-from ...cfg.cfg_utils import CFGUtils
-from ..graph_region import GraphRegion
-from ..empty_node_remover import EmptyNodeRemover
-from ..jumptable_entry_condition_rewriter import JumpTableEntryConditionRewriter
-from ..condition_processor import ConditionProcessor
-from ..region_simplifiers.cascading_cond_transformer import CascadingConditionTransformer
-from ..utils import (
+from angr.utils.graph import GraphUtils
+from angr.knowledge_plugins.cfg import IndirectJumpType
+from angr.analyses.decompiler.graph_region import GraphRegion
+from angr.analyses.decompiler.empty_node_remover import EmptyNodeRemover
+from angr.analyses.decompiler.jumptable_entry_condition_rewriter import JumpTableEntryConditionRewriter
+from angr.analyses.decompiler.condition_processor import ConditionProcessor
+from angr.analyses.decompiler.region_simplifiers.cascading_cond_transformer import CascadingConditionTransformer
+from angr.analyses.decompiler.utils import (
     extract_jump_targets,
     get_ast_subexprs,
     switch_extract_cmp_bounds,
     remove_last_statement,
-    first_nonlabel_node,
+    first_nonlabel_nonphi_node,
 )
 from .structurer_nodes import (
     SequenceNode,
@@ -67,10 +68,10 @@ class DreamStructurer(StructurerBase):
         region,
         parent_map=None,
         condition_processor=None,
-        func: Optional["Function"] = None,
-        case_entry_to_switch_head: Optional[Dict[int, int]] = None,
+        func: Function | None = None,
+        case_entry_to_switch_head: dict[int, int] | None = None,
         parent_region=None,
-        improve_structurer=True,
+        **kwargs,
     ):
         super().__init__(
             region,
@@ -79,6 +80,7 @@ class DreamStructurer(StructurerBase):
             func=func,
             case_entry_to_switch_head=case_entry_to_switch_head,
             parent_region=parent_region,
+            **kwargs,
         )
 
         self._analyze()
@@ -110,6 +112,7 @@ class DreamStructurer(StructurerBase):
 
         loop_subgraph = self._region.graph
         successors = self._region.successors
+        assert successors is not None
 
         assert len(successors) <= 1
 
@@ -213,9 +216,7 @@ class DreamStructurer(StructurerBase):
         loop_body = self._to_loop_body_sequence(loop_head, loop_subgraph, loop_successors)
 
         # create a while(true) loop with sequence node being the loop body
-        loop_node = LoopNode("while", None, loop_body, addr=loop_head.addr)
-
-        return loop_node
+        return LoopNode("while", None, loop_body, addr=loop_head.addr)
 
     def _refine_loop(self, loop_node):
         while True:
@@ -238,11 +239,8 @@ class DreamStructurer(StructurerBase):
     def _refine_loop_while(loop_node):
         if loop_node.sort == "while" and loop_node.condition is None and loop_node.sequence_node.nodes:
             # it's an endless loop
-            first_node = first_nonlabel_node(loop_node.sequence_node)
-            if type(first_node) is CodeNode:
-                inner_first_node = first_node.node
-            else:
-                inner_first_node = first_node
+            first_node = first_nonlabel_nonphi_node(loop_node.sequence_node)
+            inner_first_node = first_node.node if type(first_node) is CodeNode else first_node
             if type(inner_first_node) is ConditionalBreakNode:
                 while_cond = ConditionProcessor.simplify_condition(claripy.Not(inner_first_node.condition))
                 new_seq = loop_node.sequence_node.copy()
@@ -270,6 +268,7 @@ class DreamStructurer(StructurerBase):
 
     def _to_loop_body_sequence(self, loop_head, loop_subgraph, loop_successors):
         graph = self._region.graph_with_successors
+        assert graph is not None
         loop_region_graph = networkx.DiGraph()
 
         # TODO: Make sure the loop body has been structured
@@ -315,7 +314,7 @@ class DreamStructurer(StructurerBase):
             loop_head, loop_region_graph, successors=None, graph_with_successors=None, cyclic=False, full_graph=None
         )
         structurer = self.project.analyses[DreamStructurer].prep()(
-            region, condition_processor=self.cond_proc, func=self.function
+            region, condition_processor=self.cond_proc, func=self.function, jump_tables=self.jump_tables
         )
         seq = structurer.result
 
@@ -337,14 +336,12 @@ class DreamStructurer(StructurerBase):
                 continue
             break
 
-        seq = EmptyNodeRemover(seq).result
-
-        return seq
+        return EmptyNodeRemover(seq).result
 
     def _make_sequence(self):
         seq = SequenceNode(None)
 
-        for node in CFGUtils.quasi_topological_sort_nodes(self._region.graph):
+        for node in GraphUtils.quasi_topological_sort_nodes(self._region.graph):
             seq.add_node(CodeNode(node, self.cond_proc.reaching_conditions.get(node, None)))
 
         if seq.nodes:
@@ -377,7 +374,7 @@ class DreamStructurer(StructurerBase):
             node.condition_and_nodes = new_cond_and_nodes
 
             if node.else_node is not None:
-                node.else_node = walker._handle(node.else_node)
+                node.else_node = walker._handle(node.else_node)  # type: ignore
             return node
 
         def _handle_SwitchCaseNode(node, **kwargs):  # pylint:disable=unused-argument
@@ -469,7 +466,7 @@ class DreamStructurer(StructurerBase):
                     self._merge_nodes(node_0.node, node_1.node),
                     node_0.reaching_condition,
                 )
-                seq.nodes = seq.nodes[:i] + [new_node] + seq.nodes[i + 2 :]
+                seq.nodes = [*seq.nodes[:i], new_node, *seq.nodes[i + 2 :]]
                 continue
             i += 1
 
@@ -485,9 +482,7 @@ class DreamStructurer(StructurerBase):
         :return:        None
         """
 
-        jump_tables = self.kb.cfgs["CFGFast"].jump_tables
-
-        addr2nodes: Dict[int, Set[CodeNode]] = defaultdict(set)
+        addr2nodes: dict[int, set[CodeNode]] = defaultdict(set)
         for node in seq.nodes:
             addr2nodes[node.addr].add(node)
 
@@ -496,14 +491,14 @@ class DreamStructurer(StructurerBase):
                 node = seq.nodes[i]
 
                 # Jumptable_AddressLoadedFromMemory
-                r = self._make_switch_cases_address_loaded_from_memory(seq, i, node, addr2nodes, jump_tables)
+                r = self._make_switch_cases_address_loaded_from_memory(seq, i, node, addr2nodes)
                 if r:
                     # we found a node that looks like a switch-case. seq.nodes are changed. resume to find the next such
                     # case
                     break
 
                 # Jumptable_AddressComputed
-                r = self._make_switch_cases_address_computed(seq, i, node, addr2nodes, jump_tables)
+                r = self._make_switch_cases_address_computed(seq, i, node, addr2nodes)
                 if r:
                     break
 
@@ -511,9 +506,7 @@ class DreamStructurer(StructurerBase):
                 # we did not find any node that looks like a switch-case. exit.
                 break
 
-    def _make_switch_cases_address_loaded_from_memory(
-        self, seq, i, node, addr2nodes: Dict[int, Set[CodeNode]], jump_tables: Dict[int, IndirectJump]
-    ) -> bool:
+    def _make_switch_cases_address_loaded_from_memory(self, seq, i, node, addr2nodes: dict[int, set[CodeNode]]) -> bool:
         """
         A typical jump table involves multiple nodes, which look like the following:
 
@@ -531,6 +524,8 @@ class DreamStructurer(StructurerBase):
 
         try:
             last_stmt = self.cond_proc.get_last_statement(node)
+            if not isinstance(last_stmt, ailment.Stmt.ConditionalJump):
+                return False
         except EmptyBlockNotice:
             return False
         successor_addrs = extract_jump_targets(last_stmt)
@@ -538,14 +533,14 @@ class DreamStructurer(StructurerBase):
             return False
 
         for t in successor_addrs:
-            if t in addr2nodes and t in jump_tables:
+            if t in addr2nodes and t in self.jump_tables:
                 # this is a candidate!
                 target = t
                 break
         else:
             return False
 
-        jump_table = jump_tables[target]
+        jump_table = self.jump_tables[target]
         if jump_table.type != IndirectJumpType.Jumptable_AddressLoadedFromMemory:
             return False
 
@@ -568,16 +563,20 @@ class DreamStructurer(StructurerBase):
             return False
 
         # build switch-cases
+        assert jump_table.jumptable_entries is not None
         cases, node_default, to_remove = self._switch_build_cases(
             seq, cmp_lb, jump_table.jumptable_entries, i, node_b_addr, addr2nodes
         )
-        if node_default is None:
-            switch_end_addr = node_b_addr
-        else:
-            # we don't know what the end address of this switch-case structure is. let's figure it out
-            switch_end_addr = None
-        self._switch_handle_gotos(cases, node_default, switch_end_addr)
+        # if we don't know what the end address of this switch-case structure is, let's figure it out
+        switch_end_addr = (
+            node_b_addr
+            if node_default is None
+            else self._switch_find_switch_end_addr(cases, node_default, {nn.addr for nn in self._region.graph})
+        )
+        if switch_end_addr is not None:
+            self._switch_handle_gotos(cases, node_default, switch_end_addr)
 
+        assert last_stmt.ins_addr is not None
         self._make_switch_cases_core(
             seq,
             i,
@@ -594,12 +593,10 @@ class DreamStructurer(StructurerBase):
 
         return True
 
-    def _make_switch_cases_address_computed(
-        self, seq, i, node, addr2nodes: Dict[int, Set[CodeNode]], jump_tables: Dict[int, IndirectJump]
-    ) -> bool:
-        if node.addr not in jump_tables:
+    def _make_switch_cases_address_computed(self, seq, i, node, addr2nodes: dict[int, set[CodeNode]]) -> bool:
+        if node.addr not in self.jump_tables:
             return False
-        jump_table = jump_tables[node.addr]
+        jump_table = self.jump_tables[node.addr]
         if jump_table.type != IndirectJumpType.Jumptable_AddressComputed:
             return False
 
@@ -625,9 +622,11 @@ class DreamStructurer(StructurerBase):
         cmp_expr, cmp_lb, cmp_ub = cmp  # pylint:disable=unused-variable
 
         jumptable_entries = jump_table.jumptable_entries
+        assert jumptable_entries is not None
 
         if isinstance(last_stmt.false_target, ailment.Expr.Const):
             default_addr = last_stmt.false_target.value
+            assert isinstance(default_addr, int)
         else:
             return False
 
@@ -659,13 +658,14 @@ class DreamStructurer(StructurerBase):
         i,
         node,
         cmp_expr,
-        cases: ODict,
+        cases: OrderedDict,
         node_default,
         addr,
         addr2nodes,
         to_remove,
+        *,
+        jumptable_addr: int,
         node_a=None,
-        jumptable_addr=None,
     ):
         scnode = SwitchCaseNode(cmp_expr, cases, node_default, addr=addr)
         scnode = CodeNode(scnode, node.reaching_condition)
@@ -695,8 +695,8 @@ class DreamStructurer(StructurerBase):
         rewriter.walk(seq)  # update SequenceNodes in-place
 
     def _switch_unpack_sequence_node(
-        self, seq: SequenceNode, node_a, node_b_addr: int, jumptable, addr2nodes: Dict[int, Set[CodeNode]]
-    ) -> Tuple[bool, Optional[CodeNode]]:
+        self, seq: SequenceNode, node_a, node_b_addr: int, jumptable, addr2nodes: dict[int, set[CodeNode]]
+    ) -> tuple[bool, CodeNode | None]:
         """
         We might have already structured the actual body of the switch-case structure into a single Sequence node (node
         A). If that is the case, we un-structure the sequence node in this method.
@@ -713,19 +713,17 @@ class DreamStructurer(StructurerBase):
 
         jumptable_entries = jumptable.jumptable_entries
 
-        if isinstance(node_a.node, SequenceNode):
-            node_a_block_addrs = {n.addr for n in node_a.node.nodes}
-        else:
-            node_a_block_addrs = set()
+        node_a_block_addrs = {n.addr for n in node_a.node.nodes} if isinstance(node_a.node, SequenceNode) else set()
         #
         # if that is the case, we un-structure it here
         if all(entry_addr in addr2nodes for entry_addr in jumptable_entries):
             return True, node_a
-        elif self._switch_check_existence_of_jumptable_entries(
+        if self._switch_check_existence_of_jumptable_entries(
             jumptable_entries, node_a_block_addrs, set(addr2nodes.keys()), node_a.addr, node_b_addr
         ):
             # unpacking is needed
             for n in node_a.node.nodes:
+                assert n.addr is not None
                 if isinstance(n, ConditionNode):
                     unpacked = self._switch_unpack_condition_node(n, jumptable)
                     if unpacked is None:
@@ -751,7 +749,7 @@ class DreamStructurer(StructurerBase):
         # should have been handling it when dealing with multi-exit regions. ignore it here.
         return True, node_a
 
-    def _switch_unpack_condition_node(self, cond_node: ConditionNode, jumptable) -> Optional[CodeNode]:
+    def _switch_unpack_condition_node(self, cond_node: ConditionNode, jumptable) -> CodeNode | None:
         """
         Unpack condition nodes by only removing one condition in the form of
         <Bool jump_table_402020 == 0x402ac4>.
@@ -778,7 +776,7 @@ class DreamStructurer(StructurerBase):
                 if (
                     arg.op == "__eq__"
                     and arg.args[0] is jumptable_var
-                    and isinstance(arg.args[1], claripy.Bits)
+                    and isinstance(arg.args[1], claripy.ast.Bits)
                     and arg.args[1].concrete
                 ):
                     # found it
@@ -791,7 +789,7 @@ class DreamStructurer(StructurerBase):
                 # unsupported
                 return None
         elif cond.op == "__eq__":
-            if cond.args[0] is jumptable_var and isinstance(cond.args[1], claripy.Bits) and cond.args[1].concrete:
+            if cond.args[0] is jumptable_var and isinstance(cond.args[1], claripy.ast.Bits) and cond.args[1].concrete:
                 # found it
                 eq_condition = cond
                 true_node = cond_node.true_node
@@ -811,14 +809,15 @@ class DreamStructurer(StructurerBase):
             return None
 
         return CodeNode(
-            ConditionNode(cond_node.addr, claripy.true, remaining_cond, true_node, false_node=false_node), eq_condition
+            ConditionNode(cond_node.addr, claripy.true(), remaining_cond, true_node, false_node=false_node),
+            eq_condition,
         )
 
     def _switch_check_existence_of_jumptable_entries(
         self,
         jumptable_entries,
-        node_a_block_addrs: Set[int],
-        known_node_addrs: Set[int],
+        node_a_block_addrs: set[int],
+        known_node_addrs: set[int],
         node_a_addr: int,
         node_b_addr: int,
     ) -> bool:
@@ -846,22 +845,18 @@ class DreamStructurer(StructurerBase):
                     for _, dst, data in self.function.graph.out_edges(entry_node, data=True):
                         if data.get("type", "transition") != "call":
                             successors.append(dst)
-                    if len(successors) == 1:
+                    if len(successors) == 1 and successors[0].addr in all_node_addrs:
                         # found the single successor
-                        if successors[0].addr in all_node_addrs:
-                            expected_node_a_addrs.add(successors[0].addr)
-                            continue
+                        expected_node_a_addrs.add(successors[0].addr)
+                        continue
             # it's also possible that this is just a jump that breaks out of the switch-case. we simply ignore it.
             continue
 
         # finally, make sure all expected nodes exist
-        if node_a_block_addrs.issuperset((expected_node_a_addrs | {node_a_addr}) - {node_b_addr}):
-            return True
+        # not sure what is going on... if it's false
+        return node_a_block_addrs.issuperset((expected_node_a_addrs | {node_a_addr}) - {node_b_addr})
 
-        # not sure what is going on...
-        return False
-
-    def _switch_find_jumptable_entry_node(self, entry_addr: int, addr2nodes: Dict[int, Set[CodeNode]]) -> Optional[Any]:
+    def _switch_find_jumptable_entry_node(self, entry_addr: int, addr2nodes: dict[int, set[CodeNode]]) -> Any | None:
         """
         Find the correct node for a given jump table entry address in addr2nodes.
 
@@ -903,11 +898,11 @@ class DreamStructurer(StructurerBase):
         self,
         seq: SequenceNode,
         cmp_lb: int,
-        jumptable_entries: List[int],
+        jumptable_entries: list[int],
         head_node_idx: int,
         node_b_addr: int,
-        addr2nodes: Dict[int, Set[CodeNode]],
-    ) -> Tuple[ODict, Any, Any]:
+        addr2nodes: dict[int, set[CodeNode]],
+    ) -> tuple[OrderedDict, Any, Any]:
         """
         Discover all cases for the switch-case structure and build the switch-cases dict.
 
@@ -920,14 +915,14 @@ class DreamStructurer(StructurerBase):
         :return:                    A tuple of (dict of cases, the default node if exists, nodes to remove).
         """
 
-        cases: ODict[Union[int, Tuple[int, ...]], SequenceNode] = OrderedDict()
+        cases: OrderedDict[int | tuple[int, ...], SequenceNode] = OrderedDict()
         to_remove = set()
-        node_default = addr2nodes.get(node_b_addr, None)
+        node_default = addr2nodes.get(node_b_addr)
         if node_default is not None:
             node_default = next(iter(node_default))
 
         entry_addrs_set = set(jumptable_entries)
-        converted_nodes: Dict[int, Any] = {}
+        converted_nodes: dict[int, Any] = {}
         entry_addr_to_ids = defaultdict(set)
 
         for j, entry_addr in enumerate(jumptable_entries):
@@ -957,11 +952,11 @@ class DreamStructurer(StructurerBase):
                         )
                     ],
                 )
-                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true)])
+                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true())])
                 converted_nodes[entry_addr] = case_node
                 continue
 
-            case_node = SequenceNode(entry_node.addr, nodes=[CodeNode(entry_node.node, claripy.true)])
+            case_node = SequenceNode(entry_node.addr, nodes=[CodeNode(entry_node.node, claripy.true())])
             to_remove.add(entry_node)
             entry_node_idx = seq.nodes.index(entry_node)
 
@@ -981,7 +976,7 @@ class DreamStructurer(StructurerBase):
                         )
                     ],
                 )
-                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true)])
+                case_node = SequenceNode(0, nodes=[CodeNode(case_inner_node, claripy.true())])
                 converted_nodes[entry_addr] = case_node
                 continue
 
@@ -1001,9 +996,9 @@ class DreamStructurer(StructurerBase):
                 for node_ in sorted_guarded_nodes:
                     if node_ is not entry_node and node_.addr not in entry_addrs_set:
                         # fix reaching condition
-                        reaching_condition_subexprs = {
-                            ex for ex in get_ast_subexprs(node_.reaching_condition)
-                        }.difference(set(cond_subexprs))
+                        reaching_condition_subexprs = set(get_ast_subexprs(node_.reaching_condition)).difference(
+                            set(cond_subexprs)
+                        )
                         new_reaching_condition = claripy.And(*reaching_condition_subexprs)
                         new_node = CodeNode(node_.node, new_reaching_condition)
                         case_node.add_node(new_node)
@@ -1112,7 +1107,7 @@ class DreamStructurer(StructurerBase):
     def _nodes_guarded_by_common_subexpr(seq, common_subexpr, starting_idx):
         candidates = []
 
-        if common_subexpr is claripy.true:
+        if common_subexpr is claripy.true():
             return []
         for j, node_1 in enumerate(seq.nodes[starting_idx:]):
             rcond_1 = getattr(node_1, "reaching_condition", None)
@@ -1138,8 +1133,7 @@ class DreamStructurer(StructurerBase):
             )
             new_nodes.append(new_node)
 
-        new_node = SequenceNode(None if not new_nodes else new_nodes[0].addr, nodes=new_nodes)
-        return new_node
+        return SequenceNode(None if not new_nodes else new_nodes[0].addr, nodes=new_nodes)
 
     def _replace_complex_reaching_conditions(self, seq: SequenceNode):
         for i in range(len(seq.nodes)):
@@ -1220,4 +1214,4 @@ class DreamStructurer(StructurerBase):
 
 
 # delayed import
-from ..sequence_walker import SequenceWalker  # pylint:disable=wrong-import-position
+from angr.analyses.decompiler.sequence_walker import SequenceWalker  # pylint:disable=wrong-import-position
