@@ -1,29 +1,29 @@
+from __future__ import annotations
+
+import contextlib
 import logging
 from collections import defaultdict
-from typing import Union, Optional, Sequence, Tuple, Any
+from collections.abc import Sequence
+from typing import Any
 
-import pyvex
 import archinfo
-from angr.knowledge_plugins import Function
+import pypcode
+import pyvex
 
 from . import Analysis
-
-from ..utils.library import get_cpp_function_name
-from ..utils.formatting import ansi_color_enabled, ansi_color, add_edge_to_buffer
-from ..block import DisassemblerInsn, CapstoneInsn, SootBlockNode
-from ..codenode import BlockNode
+from angr.analyses import AnalysesHub
+from angr.engines import pcode
+from angr.errors import AngrTypeError
+from angr.knowledge_plugins import Function
+from angr.utils.library import get_cpp_function_name
+from angr.utils.formatting import ansi_color_enabled, ansi_color, add_edge_to_buffer
+from angr.block import DisassemblerInsn, CapstoneInsn, SootBlockNode
+from angr.codenode import BlockNode
 from .disassembly_utils import decode_instruction
 
-try:
-    from ..engines import pcode
-    import pypcode
 
-    IRSBType = Union[pyvex.IRSB, pcode.lifter.IRSB]
-    IROpObjType = Union[pyvex.stmt.IRStmt, pypcode.PcodeOp]
-except ImportError:
-    pcode = None
-    IRSBType = pyvex.IRSB
-    IROpObjType = pyvex.stmt
+IRSBType = pyvex.IRSB | pcode.lifter.IRSB
+IROpObjType = pyvex.stmt.IRStmt | pypcode.PcodeOp
 
 l = logging.getLogger(name=__name__)
 
@@ -38,8 +38,7 @@ class DisassemblyPiece:
         x = self._render(formatting)
         if len(x) == 1:
             return [self.highlight(x[0], formatting)]
-        else:
-            return x
+        return x
 
     def _render(self, formatting):
         raise NotImplementedError
@@ -116,9 +115,9 @@ class Label(DisassemblyPiece):
 class IROp(DisassemblyPiece):
     __slots__ = (
         "addr",
-        "seq",
-        "obj",
         "irsb",
+        "obj",
+        "seq",
     )
 
     addr: int
@@ -133,6 +132,8 @@ class IROp(DisassemblyPiece):
         self.irsb = irsb
 
     def __str__(self):
+        if pcode and isinstance(self.obj, pypcode.PcodeOp):
+            return pypcode.PcodePrettyPrinter.fmt_op(self.obj)
         return str(self.obj)
 
     def _render(self, formatting):  # pylint:disable=unused-argument
@@ -214,15 +215,19 @@ class Instruction(DisassemblyPiece):
         assert hasattr(self.insn, "operands")
 
         op_str = self.insn.op_str
-        # NOTE: the size of dummy_operands is not necessarily equal to
-        # operands count of capstone disasm result. If we check
-        # len(self.operands) == len(self.insn.operands)
-        # special cases in arm disassembly will mess up the code
         dummy_operands = self.split_arm_op_string(op_str)
 
         for operand in dummy_operands:
             opr_pieces = self.split_op_string(operand)
             cur_operand = []
+
+            if not (operand and opr_pieces):
+                # opr_pieces may contain empty string when invalid disasm
+                # result is generated by capstone
+                l.error(f'Failed to parse insn "{self.insn}". Please report.')
+                self.operands.clear()
+                break
+
             if opr_pieces[0][0].isalpha() and opr_pieces[0] in self.arch.registers:
                 cur_operand.append(Register(opr_pieces[0]))
                 # handle register's suffix (e.g. "sp!", "d0[1]", "v0.16b")
@@ -264,30 +269,26 @@ class Instruction(DisassemblyPiece):
             self.operands.append(cur_operand)
 
         for i, opr in enumerate(self.operands):
-            if i < len(self.insn.operands):
-                op_type = self.insn.operands[i].type
-            else:
-                # set extra dummy operand type to default 0
-                op_type = 0
+            # set extra dummy operand type to default 0
+            op_type = self.insn.operands[i].type if i < len(self.insn.operands) else 0
             self.operands[i] = Operand.build(op_type, i, opr, self)
 
         if len(self.operands) == 0 and len(self.insn.operands) != 0:
             l.error("Operand parsing failed for instruction %s at address %x", str(self.insn), self.insn.address)
-            self.operands = []
             return
 
     @staticmethod
     def split_arm_op_string(op_str: str):
-        # Split arm operand string by comma outside brackets
+        # Split arm operand string with commas outside the square brackets
         pieces = []
-        outside_brackets = True
+        in_square_brackets = False
         cur_opr = ""
         for c in op_str:
-            if c == "[" or c == "{":
-                outside_brackets = False
-            if c == "]" or c == "}":
-                outside_brackets = True
-            if c == "," and outside_brackets:
+            if c == "[":
+                in_square_brackets = True
+            if c == "]":
+                in_square_brackets = False
+            if c == "," and not in_square_brackets:
                 pieces.append(cur_opr)
                 cur_opr = ""
                 continue
@@ -439,7 +440,7 @@ class SootExpressionTarget(SootExpression):
         self.target_stmt_idx = target_stmt_idx
 
     def _render(self, formatting=None):
-        return ["Goto %d" % self.target_stmt_idx]
+        return [f"Goto {self.target_stmt_idx}"]
 
 
 class SootExpressionStaticFieldRef(SootExpression):
@@ -489,22 +490,21 @@ class SootStatement(DisassemblyPiece):
         return self.addr.stmt_idx
 
     def _parse(self):
-        func = "_parse_%s" % self.raw_stmt.__class__.__name__
+        func = f"_parse_{self.raw_stmt.__class__.__name__}"
 
         if hasattr(self, func):
             getattr(self, func)()
         else:
             # print func
-            self.components += ["NotImplemented: %s" % func]
+            self.components += [f"NotImplemented: {func}"]
 
     def _expr(self, expr):
-        func = "_handle_%s" % expr.__class__.__name__
+        func = f"_handle_{expr.__class__.__name__}"
 
         if hasattr(self, func):
             return getattr(self, func)(expr)
-        else:
-            # print func
-            return SootExpression(str(expr))
+        # print func
+        return SootExpression(str(expr))
 
     def _render(self, formatting=None):
         return [
@@ -632,9 +632,9 @@ class Operand(DisassemblyPiece):
             70: Operand,  # ARM64 BARRIER
         }
 
-        cls = MAPPING.get(operand_type, None)
+        cls = MAPPING.get(operand_type)
         if cls is None:
-            raise ValueError("Unknown capstone operand type %s." % operand_type)
+            raise ValueError(f"Unknown capstone operand type {operand_type}.")
 
         operand = cls(op_num, children, parentinsn)
 
@@ -645,7 +645,7 @@ class Operand(DisassemblyPiece):
                 # Indirect addressing in x86_64
                 # 400520  push [rip+0x200782] ==>  400520  push [0x600ca8]
                 absolute_addr = parentinsn.addr + parentinsn.size + op1.val
-                return MemoryOperand(1, operand.prefix + ["[", Value(absolute_addr, False), "]"], parentinsn)
+                return MemoryOperand(1, [*operand.prefix, "[", Value(absolute_addr, False), "]"], parentinsn)
 
         return operand
 
@@ -662,15 +662,12 @@ class RegisterOperand(Operand):
     def _render(self, formatting):
         custom_value_str = None
         if formatting is not None:
-            try:
+            with contextlib.suppress(KeyError):
                 custom_value_str = formatting["custom_values_str"][self.ident]
-            except KeyError:
-                pass
 
         if custom_value_str:
             return [custom_value_str]
-        else:
-            return super()._render(formatting)
+        return super()._render(formatting)
 
 
 class MemoryOperand(Operand):
@@ -682,7 +679,7 @@ class MemoryOperand(Operand):
         # or
         # [ '[', Register, ']' ]
         # or
-        # [ Value, '(', Regsiter, ')' ]
+        # [ Value, '(', Register, ')' ]
 
         # it will be converted into more meaningful and Pythonic properties
 
@@ -708,7 +705,7 @@ class MemoryOperand(Operand):
             elif "(" in self.children:
                 self._parse_memop_paren()
             else:
-                raise ValueError()
+                raise ValueError
 
         except ValueError:
             l.error("Failed to parse operand children %s. Please report to Fish.", self.children)
@@ -746,7 +743,7 @@ class MemoryOperand(Operand):
                 self.suffix_str = "!"
                 close_square_pos -= 1
             else:
-                raise ValueError()
+                raise ValueError
 
         self.values = self.children[square_bracket_pos + 1 : close_square_pos]
 
@@ -782,43 +779,54 @@ class MemoryOperand(Operand):
         if self.prefix is None:
             # we failed in parsing. use the default rendering
             return super()._render(formatting)
+        values_style = self.values_style
+        show_prefix = True
+        custom_values_str = None
+
+        if formatting is not None:
+            with contextlib.suppress(KeyError):
+                values_style = formatting["values_style"][self.ident]
+
+            try:
+                show_prefix_str = formatting["show_prefix"][self.ident]
+                if show_prefix_str in ("false", "False"):
+                    show_prefix = False
+            except KeyError:
+                pass
+
+            with contextlib.suppress(KeyError):
+                custom_values_str = formatting["custom_values_str"][self.ident]
+
+        prefix_str = " ".join(self.prefix) + " " if show_prefix and self.prefix else ""
+        if custom_values_str is not None:
+            value_str = custom_values_str
         else:
-            values_style = self.values_style
-            show_prefix = True
-            custom_values_str = None
+            value_str = "".join(x.render(formatting)[0] if not isinstance(x, (bytes, str)) else x for x in self.values)
 
-            if formatting is not None:
-                try:
-                    values_style = formatting["values_style"][self.ident]
-                except KeyError:
-                    pass
+        if values_style == "curly":
+            left_paren, right_paren = "{", "}"
+        elif values_style == "paren":
+            left_paren, right_paren = "(", ")"
+        else:  # square
+            left_paren, right_paren = "[", "]"
 
-                try:
-                    show_prefix_str = formatting["show_prefix"][self.ident]
-                    if show_prefix_str in ("false", "False"):
-                        show_prefix = False
-                except KeyError:
-                    pass
+        if self.offset:
+            offset_str = "".join(x.render(formatting)[0] if not isinstance(x, (bytes, str)) else x for x in self.offset)
 
-                try:
-                    custom_values_str = formatting["custom_values_str"][self.ident]
-                except KeyError:
-                    pass
+            # combine values and offsets according to self.offset_location
+            if self.offset_location == "prefix":
+                value_str = f"{offset_str}{left_paren}{value_str}{right_paren}"
+            elif self.offset_location == "before_value":
+                value_str = f"{left_paren}{offset_str}{value_str}{right_paren}"
+            else:  # after_value
+                value_str = f"{left_paren}{value_str}{offset_str}{right_paren}"
+        else:
+            value_str = left_paren + value_str + right_paren
 
-            prefix_str = " ".join(self.prefix) + " " if show_prefix and self.prefix else ""
-            if custom_values_str is not None:
-                value_str = custom_values_str
-            else:
-                value_str = "".join(
-                    x.render(formatting)[0] if not isinstance(x, (bytes, str)) else x for x in self.values
-                )
+        segment_selector_str = "" if self.segment_selector is None else self.segment_selector
 
-            if values_style == "curly":
-                left_paren, right_paren = "{", "}"
-            elif values_style == "paren":
-                left_paren, right_paren = "(", ")"
-            else:  # square
-                left_paren, right_paren = "[", "]"
+        if segment_selector_str and prefix_str:
+            prefix_str += " "
 
             if self.offset:
                 offset_str = "".join(
@@ -840,7 +848,7 @@ class MemoryOperand(Operand):
             if segment_selector_str and prefix_str:
                 prefix_str += " "
 
-            return [f"{prefix_str}{segment_selector_str}{value_str}{self.suffix_str}"]
+        return [f"{prefix_str}{segment_selector_str}{value_str}{self.suffix_str}"]
 
 
 class OperandPiece(DisassemblyPiece):  # pylint: disable=abstract-method
@@ -882,15 +890,13 @@ class Value(OperandPiece):
                 style = formatting["int_styles"][self.ident]
                 if style[0] == "hex":
                     if self.render_with_sign:
-                        return ["%#+x" % self.val]
-                    else:
-                        return ["%#x" % self.val]
-                elif style[0] == "dec":
+                        return [f"{self.val:+#x}"]
+                    return [f"{self.val:#x}"]
+                if style[0] == "dec":
                     if self.render_with_sign:
-                        return ["%+d" % self.val]
-                    else:
-                        return [str(self.val)]
-                elif style[0] == "label":
+                        return [f"{self.val:+d}"]
+                    return [str(self.val)]
+                if style[0] == "label":
                     labeloffset = style[1]
                     if labeloffset == 0:
                         lbl = self.project.kb.labels[self.val]
@@ -913,20 +919,17 @@ class Value(OperandPiece):
 
         if self.val in self.project.kb.labels:
             lbl = self.project.kb.labels[self.val]
-            if func is not None:
+            if func is not None and lbl == func.name and func.name != func.demangled_name:
                 # see if lbl == func.name and func.demangled_name != func.name. if so, we prioritize the
                 # demangled name
-                if lbl == func.name and func.name != func.demangled_name:
-                    normalized_name = get_cpp_function_name(func.demangled_name, specialized=False, qualified=True)
-                    return [normalized_name]
+                normalized_name = get_cpp_function_name(func.demangled_name)
+                return [normalized_name]
             return [("+" if self.render_with_sign else "") + lbl]
-        elif func is not None:
+        if func is not None:
             return [func.demangled_name]
-        else:
-            if self.render_with_sign:
-                return ["%#+x" % self.val]
-            else:
-                return ["%#x" % self.val]
+        if self.render_with_sign:
+            return [f"{self.val:+#x}"]
+        return [f"{self.val:#x}"]
 
 
 class Comment(DisassemblyPiece):
@@ -957,11 +960,11 @@ class Disassembly(Analysis):
 
     def __init__(
         self,
-        function: Optional[Function] = None,
-        ranges: Optional[Sequence[Tuple[int, int]]] = None,
+        function: Function | None = None,
+        ranges: Sequence[tuple[int, int]] | None = None,
         thumb: bool = False,
         include_ir: bool = False,
-        block_bytes: Optional[bytes] = None,
+        block_bytes: bytes | None = None,
     ):
         self.raw_result = []
         self.raw_result_map = {
@@ -987,7 +990,7 @@ class Disassembly(Analysis):
         elif ranges is not None:
             cfg = self.project.kb.cfgs.get_most_accurate()
             fallback = True
-            if cfg is not None:
+            if self._block_bytes is None and cfg is not None:
                 try:
                     self._graph = cfg.graph
                     for start, end in ranges:
@@ -1075,9 +1078,12 @@ class Disassembly(Analysis):
 
         if irsb.statements is not None:
             if pcode is not None and isinstance(self.project.factory.default_engine, pcode.HeavyPcodeMixin):
-                for ins in irsb._instructions:
-                    addr = ins.address.offset
-                    addr_to_ops_map[addr].extend([IROp(addr, op.seq.uniq, op, irsb) for op in ins.ops])
+                addr = None
+                for stmt_idx, op in enumerate(irsb._ops):
+                    if op.opcode == pypcode.OpCode.IMARK:
+                        addr = op.inputs[0].offset
+                    else:
+                        addr_to_ops_map[addr].append(IROp(addr, stmt_idx, op, irsb))
             else:
                 for seq, stmt in enumerate(irsb.statements):
                     if isinstance(stmt, pyvex.stmt.IMark):
@@ -1111,7 +1117,7 @@ class Disassembly(Analysis):
                 aligned_block_addr = block.addr
                 cs = self.project.arch.capstone
             if block.bytestr is None:
-                bytestr = self.project.loader.memory.load(aligned_block_addr, block.size)
+                bytestr = self.project.factory.block(aligned_block_addr, block.size).bytes
             else:
                 bytestr = block.bytestr
             self.block_to_insn_addrs[block.addr] = []
@@ -1131,7 +1137,9 @@ class Disassembly(Analysis):
                 self.raw_result_map["instructions"][stmt.addr] = stmt
                 self.block_to_insn_addrs[block.addr].append(stmt.addr)
         else:
-            raise TypeError("")
+            raise AngrTypeError(
+                f"Cannot disassemble block with architecture {self.project.arch} for block type {type(block)}"
+            )
 
         if self._include_ir:
             b = self.project.factory.block(block.addr, size=block.size)
@@ -1143,8 +1151,9 @@ class Disassembly(Analysis):
         show_edges: bool = True,
         show_addresses: bool = True,
         show_bytes: bool = False,
-        ascii_only: Optional[bool] = None,
+        ascii_only: bool | None = None,
         color: bool = True,
+        min_edge_depth: int = 0,
     ) -> str:
         """
         Render the disassembly to a string, with optional edges and addresses.
@@ -1158,22 +1167,24 @@ class Disassembly(Analysis):
 
         if formatting is None:
             formatting = {
-                "colors": {
-                    "address": "gray",
-                    "bytes": "cyan",
-                    "edge": "yellow",
-                    Label: "bright_yellow",
-                    ConstantOperand: "cyan",
-                    MemoryOperand: "yellow",
-                    Comment: "gray",
-                    Hook: "green",
-                }
-                if ansi_color_enabled and color
-                else {},
+                "colors": (
+                    {
+                        "address": "gray",
+                        "bytes": "cyan",
+                        "edge": "yellow",
+                        Label: "bright_yellow",
+                        ConstantOperand: "cyan",
+                        MemoryOperand: "yellow",
+                        Comment: "gray",
+                        Hook: "green",
+                    }
+                    if ansi_color_enabled and color
+                    else {}
+                ),
                 "format_callback": lambda item, s: ansi_color(s, formatting["colors"].get(type(item), None)),
             }
 
-        def col(item: Any) -> Optional[str]:
+        def col(item: Any) -> str | None:
             try:
                 return formatting["colors"][item]
             except KeyError:
@@ -1272,7 +1283,7 @@ class Disassembly(Analysis):
             for f, t in sorted(edges_by_line, key=lambda e: abs(e[0] - e[1])):
                 add_edge_to_buffer(edge_buf, ref_buf, f, t, lambda s: ansi_color(s, edge_col), ascii_only=ascii_only)
                 add_edge_to_buffer(ref_buf, ref_buf, f, t, ascii_only=ascii_only)
-            max_edge_depth = max(map(len, ref_buf))
+            max_edge_depth = max(*map(len, ref_buf), min_edge_depth)
 
             # Justify edge and combine with disassembly
             for i, line in enumerate(buf):
@@ -1281,6 +1292,4 @@ class Disassembly(Analysis):
         return "\n".join(buf)
 
 
-from angr.analyses import AnalysesHub
-
 AnalysesHub.register_default("Disassembly", Disassembly)
