@@ -1,15 +1,16 @@
-from typing import Optional, Set, List, Tuple, TYPE_CHECKING, Dict
+from __future__ import annotations
+from typing import TYPE_CHECKING
 import logging
 
 import networkx
 
-import ailment
+import angr.ailment as ailment
+from angr.ailment import AILBlockWalker
 
-from ..codenode import BlockNode
-from ..sim_variable import SimMemoryVariable
-from ..knowledge_plugins.functions import Function
+from angr.codenode import BlockNode
+from angr.sim_variable import SimMemoryVariable
+from angr.knowledge_plugins.functions import Function
 from .analysis import Analysis, AnalysesHub
-from .decompiler.ailblock_walker import AILBlockWalker
 
 if TYPE_CHECKING:
     from angr.analyses.decompiler.decompiler import Decompiler
@@ -38,7 +39,7 @@ class BaseProxiNode:
     Base class for all nodes in a proximity graph.
     """
 
-    def __init__(self, type_: int, ref_at: Optional[Set[int]] = None):
+    def __init__(self, type_: int, ref_at: set[int] | None = None):
         self.type_ = type_
         self.ref_at = ref_at
 
@@ -54,7 +55,7 @@ class FunctionProxiNode(BaseProxiNode):
     Proximity node showing current and expanded function calls in graph.
     """
 
-    def __init__(self, func, ref_at: Optional[Set[int]] = None):
+    def __init__(self, func, ref_at: set[int] | None = None):
         super().__init__(ProxiNodeTypes.Function, ref_at=ref_at)
         self.func = func
 
@@ -70,7 +71,7 @@ class VariableProxiNode(BaseProxiNode):
     Variable arg node
     """
 
-    def __init__(self, addr, name, ref_at: Optional[Set[int]] = None):
+    def __init__(self, addr, name, ref_at: set[int] | None = None):
         super().__init__(ProxiNodeTypes.Variable, ref_at=ref_at)
         self.addr = addr
         self.name = name
@@ -87,7 +88,7 @@ class StringProxiNode(BaseProxiNode):
     String arg node
     """
 
-    def __init__(self, addr, content, ref_at: Optional[Set[int]] = None):
+    def __init__(self, addr, content, ref_at: set[int] | None = None):
         super().__init__(ProxiNodeTypes.String, ref_at=ref_at)
         self.addr = addr
         self.content = content
@@ -104,7 +105,7 @@ class CallProxiNode(BaseProxiNode):
     Call node
     """
 
-    def __init__(self, callee, ref_at: Optional[Set[int]] = None, args: Optional[Tuple[BaseProxiNode]] = None):
+    def __init__(self, callee, ref_at: set[int] | None = None, args: tuple[BaseProxiNode] | None = None):
         super().__init__(ProxiNodeTypes.FunctionCall, ref_at=ref_at)
         self.callee = callee
         self.args = args
@@ -127,7 +128,7 @@ class IntegerProxiNode(BaseProxiNode):
     Int arg node
     """
 
-    def __init__(self, value: int, ref_at: Optional[Set[int]] = None):
+    def __init__(self, value: int, ref_at: set[int] | None = None):
         super().__init__(ProxiNodeTypes.Integer, ref_at=ref_at)
         self.value = value
 
@@ -163,11 +164,11 @@ class ProximityGraphAnalysis(Analysis):
 
     def __init__(
         self,
-        func: "Function",
-        cfg_model: "CFGModel",
-        xrefs: "XRefManager",
-        decompilation: Optional["Decompiler"] = None,
-        expand_funcs: Optional[Set[int]] = None,
+        func: Function,
+        cfg_model: CFGModel,
+        xrefs: XRefManager,
+        decompilation: Decompiler | None = None,
+        expand_funcs: set[int] | None = None,
     ):
         self._function = func
         self._cfg_model = cfg_model
@@ -175,11 +176,38 @@ class ProximityGraphAnalysis(Analysis):
         self._decompilation = decompilation
         self._expand_funcs = expand_funcs.copy() if expand_funcs else None
 
-        self.graph: Optional[networkx.DiGraph] = None
+        self.graph: networkx.DiGraph | None = None
         self.handled_stmts = []
 
         self._work()
 
+    def _condense_blank_nodes(self, graph: networkx.DiGraph) -> None:
+        nodes = list(graph.nodes)
+        blank_nodes: list[BaseProxiNode] = []
+
+        for node in nodes:
+            if isinstance(node, BaseProxiNode) and node.type_ == ProxiNodeTypes.Empty:
+                blank_nodes.append(node)
+            else:
+                if blank_nodes:
+                    self._merge_nodes(graph, blank_nodes)
+                    blank_nodes = []
+
+        if blank_nodes:
+            self._merge_nodes(graph, blank_nodes)
+
+    def _merge_nodes(self, graph: networkx.DiGraph, nodes: list[BaseProxiNode]) -> None:
+        for node in nodes:
+            predecessors = set(graph.predecessors(node))
+            successors = set(graph.successors(node))
+
+            for pred in predecessors:
+                for succ in successors:
+                    edge_data = graph.get_edge_data(pred, node) or {}
+                    graph.add_edge(pred, succ, **edge_data)
+
+            graph.remove_node(node)
+
     def _work(self):
         self.graph = networkx.DiGraph()
 
@@ -210,7 +238,10 @@ class ProximityGraphAnalysis(Analysis):
             self.graph.add_nodes_from(subgraph.nodes())
             self.graph.add_edges_from(subgraph.edges())
 
-    def _endnode_connector(self, func: "Function", subgraph: networkx.DiGraph):
+        # condense blank nodes after the graph has been constructed
+        self._condense_blank_nodes(self.graph)
+
+    def _endnode_connector(self, func: Function, subgraph: networkx.DiGraph):
         """
         Properly connect expanded function call's to proximity graph.
         """
@@ -238,10 +269,10 @@ class ProximityGraphAnalysis(Analysis):
                     subgraph.add_edge(end_node, succ)
 
     def _process_function(
-        self, func: "Function", graph: networkx.DiGraph, func_proxi_node: Optional[FunctionProxiNode] = None
-    ) -> List[FunctionProxiNode]:
-        to_expand: List[FunctionProxiNode] = []
-        found_blocks: Dict[BlockNode:BaseProxiNode] = {}
+        self, func: Function, graph: networkx.DiGraph, func_proxi_node: FunctionProxiNode | None = None
+    ) -> list[FunctionProxiNode]:
+        to_expand: list[FunctionProxiNode] = []
+        found_blocks: dict[BlockNode:BaseProxiNode] = {}
 
         # function calls
         for n_ in func.nodes:
@@ -259,10 +290,7 @@ class ProximityGraphAnalysis(Analysis):
                         found_blocks[block] = node
 
         # subgraph check - do before in case of recursion
-        if self.graph == graph:
-            subgraph = False
-        else:
-            subgraph = True
+        subgraph = self.graph != graph
 
         for edge in func.graph.edges:
             nodes = ()
@@ -316,18 +344,18 @@ class ProximityGraphAnalysis(Analysis):
             args.append(UnknownProxiNode("_"))
 
     def _process_decompilation(
-        self, graph: networkx.DiGraph, decompilation: "Decompiler", func_proxi_node: Optional[FunctionProxiNode] = None
-    ) -> List[FunctionProxiNode]:
-        to_expand: List[FunctionProxiNode] = []
+        self, graph: networkx.DiGraph, decompilation: Decompiler, func_proxi_node: FunctionProxiNode | None = None
+    ) -> list[FunctionProxiNode]:
+        to_expand: list[FunctionProxiNode] = []
 
         # dedup
-        string_refs: Set[int] = set()
+        string_refs: set[int] = set()
 
         # Walk the clinic structure to dump string references and function calls
         ail_graph = decompilation.clinic.cc_graph
 
         def _handle_Call(
-            stmt_idx: int, stmt: ailment.Stmt.Call, block: Optional[ailment.Block]  # pylint:disable=unused-argument
+            stmt_idx: int, stmt: ailment.Stmt.Call, block: ailment.Block | None  # pylint:disable=unused-argument
         ):  # pylint:disable=unused-argument
             func_node = self.kb.functions[stmt.target.value]
             ref_at = {stmt.ins_addr}
@@ -356,15 +384,12 @@ class ProximityGraphAnalysis(Analysis):
             expr: ailment.Stmt.Call,
             stmt_idx: int,  # pylint:disable=unused-argument
             stmt: ailment.Stmt.Statement,  # pylint:disable=unused-argument
-            block: Optional[ailment.Block],
+            block: ailment.Block | None,
         ):  # pylint:disable=unused-argument
             _handle_Call(stmt_idx, expr, block)
 
         # subgraph check - do before in case of recursion
-        if self.graph == graph:
-            subgraph = False
-        else:
-            subgraph = True
+        subgraph = self.graph != graph
 
         # Keep all default handlers, but overwrite necessary ones:
         bw = AILBlockWalker()
@@ -388,10 +413,7 @@ class ProximityGraphAnalysis(Analysis):
                             graph.add_edge(self.handled_stmts[idx - 1], current)
                     # If first block in edge, LAST handled node -> next node
                     # Else (second block in edge), prev node -> FIRST handled node
-                    if block == ail_edge[0]:
-                        proxi_node = self.handled_stmts[-1]
-                    else:
-                        proxi_node = self.handled_stmts[0]
+                    proxi_node = self.handled_stmts[-1] if block == ail_edge[0] else self.handled_stmts[0]
 
                     # Clear the handled stmts for next block
                     self.handled_stmts = []
