@@ -1,25 +1,33 @@
 # pylint:disable=missing-class-docstring
-from typing import Dict, Any, Optional, TYPE_CHECKING
+from __future__ import annotations
+from typing import Any, Union, TYPE_CHECKING
+
+from collections.abc import Sequence, Iterable
 from itertools import count
 
-from ...utils.cowdict import ChainMapCOW
+from angr.utils.constants import MAX_POINTSTO_BITS
+from .variance import Variance
 
 if TYPE_CHECKING:
     from angr.sim_variable import SimVariable
+    from .typeconsts import TypeConstant
 
 
 # Type variables and constraints
 
+TypeType = Union["TypeConstant", "TypeVariable", "DerivedTypeVariable"]
+
 
 class TypeConstraint:
     __slots__ = ()
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
-        raise NotImplementedError()
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
+        raise NotImplementedError
 
 
 class Equivalence(TypeConstraint):
     __slots__ = (
+        "_cached_hash",
         "type_a",
         "type_b",
     )
@@ -27,8 +35,9 @@ class Equivalence(TypeConstraint):
     def __init__(self, type_a, type_b):
         self.type_a = type_a
         self.type_b = type_b
+        self._cached_hash = hash((Equivalence, self.type_a, self.type_b))
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
         return f"{self.type_a.pp_str(mapping)} == {self.type_b.pp_str(mapping)}"
 
     def __repr__(self):
@@ -36,56 +45,56 @@ class Equivalence(TypeConstraint):
 
     def __eq__(self, other):
         return type(other) is Equivalence and (
-            self.type_a == other.type_a
-            and self.type_b == other.type_b
-            or self.type_b == other.type_a
-            and self.type_a == other.type_b
+            (self.type_a == other.type_a and self.type_b == other.type_b)
+            or (self.type_b == other.type_a and self.type_a == other.type_b)
         )
 
     def __hash__(self):
-        return hash((Equivalence, tuple(sorted((hash(self.type_a), hash(self.type_b))))))
+        return self._cached_hash
 
 
 class Existence(TypeConstraint):
-    __slots__ = ("type_",)
+    __slots__ = ("_cached_hash", "type_")
 
     def __init__(self, type_):
         self.type_ = type_
+        self._cached_hash = hash((Existence, self.type_))
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
         return f"V {self.type_.pp_str(mapping)}"
 
     def __repr__(self):
-        return "V %s" % self.type_
+        return f"V {self.type_}"
 
     def __eq__(self, other):
         return type(other) is Existence and self.type_ == other.type_
 
     def __hash__(self):
-        return hash((Existence, self.type_))
+        return self._cached_hash
 
     def replace(self, replacements):
         if self.type_ in replacements:
             return True, Existence(replacements[self.type_])
 
-        else:
-            replaced, new_type = self.type_.replace(replacements)
-            if replaced:
-                return True, Existence(new_type)
-            return False, self
+        replaced, new_type = self.type_.replace(replacements)
+        if replaced:
+            return True, Existence(new_type)
+        return False, self
 
 
 class Subtype(TypeConstraint):
     __slots__ = (
-        "super_type",
+        "_cached_hash",
         "sub_type",
+        "super_type",
     )
 
-    def __init__(self, sub_type, super_type):
+    def __init__(self, sub_type: TypeType, super_type: TypeType):
         self.super_type = super_type
         self.sub_type = sub_type
+        self._cached_hash = hash((Subtype, self.sub_type, self.super_type))
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
         return f"{self.sub_type.pp_str(mapping)} <: {self.super_type.pp_str(mapping)}"
 
     def __repr__(self):
@@ -95,7 +104,7 @@ class Subtype(TypeConstraint):
         return type(other) is Subtype and self.sub_type == other.sub_type and self.super_type == other.super_type
 
     def __hash__(self):
-        return hash((Subtype, hash(self.sub_type), hash(self.super_type)))
+        return self._cached_hash
 
     def replace(self, replacements):
         subtype, supertype = None, None
@@ -122,8 +131,7 @@ class Subtype(TypeConstraint):
                 subtype if subtype is not None else self.sub_type,
                 supertype if supertype is not None else self.super_type,
             )
-        else:
-            return False, self
+        return False, self
 
 
 class Add(TypeConstraint):
@@ -132,6 +140,7 @@ class Add(TypeConstraint):
     """
 
     __slots__ = (
+        "_cached_hash",
         "type_0",
         "type_1",
         "type_r",
@@ -141,13 +150,10 @@ class Add(TypeConstraint):
         self.type_0 = type_0
         self.type_1 = type_1
         self.type_r = type_r
+        self._cached_hash = hash((Add, self.type_0, self.type_1, self.type_r))
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
-        return "{} == {} + {}".format(
-            self.type_r.pp_str(mapping),
-            self.type_0.pp_str(mapping),
-            self.type_1.pp_str(mapping),
-        )
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
+        return f"{self.type_r.pp_str(mapping)} == {self.type_0.pp_str(mapping)} + {self.type_1.pp_str(mapping)}"
 
     def __repr__(self):
         return f"{self.type_r!r} == {self.type_0!r} + {self.type_1!r}"
@@ -161,7 +167,7 @@ class Add(TypeConstraint):
         )
 
     def __hash__(self):
-        return hash((Add, self.type_0, self.type_1, self.type_r))
+        return self._cached_hash
 
     def replace(self, replacements):
         t0, t1, tr = None, None, None
@@ -194,8 +200,7 @@ class Add(TypeConstraint):
                 t1 if t1 is not None else self.type_1,
                 tr if tr is not None else self.type_r,
             )
-        else:
-            return False, self
+        return False, self
 
 
 class Sub(TypeConstraint):
@@ -204,6 +209,7 @@ class Sub(TypeConstraint):
     """
 
     __slots__ = (
+        "_cached_hash",
         "type_0",
         "type_1",
         "type_r",
@@ -213,13 +219,10 @@ class Sub(TypeConstraint):
         self.type_0 = type_0
         self.type_1 = type_1
         self.type_r = type_r
+        self._cached_hash = hash((Sub, self.type_0, self.type_1, self.type_r))
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
-        return "{} == {} - {}".format(
-            self.type_r.pp_str(mapping),
-            self.type_0.pp_str(mapping),
-            self.type_1.pp_str(mapping),
-        )
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
+        return f"{self.type_r.pp_str(mapping)} == {self.type_0.pp_str(mapping)} - {self.type_1.pp_str(mapping)}"
 
     def __repr__(self):
         return f"{self.type_r!r} == {self.type_0!r} - {self.type_1!r}"
@@ -233,7 +236,7 @@ class Sub(TypeConstraint):
         )
 
     def __hash__(self):
-        return hash((Sub, self.type_0, self.type_1, self.type_r))
+        return self._cached_hash
 
     def replace(self, replacements):
         t0, t1, tr = None, None, None
@@ -266,60 +269,115 @@ class Sub(TypeConstraint):
                 t1 if t1 is not None else self.type_1,
                 tr if tr is not None else self.type_r,
             )
-        else:
-            return False, self
+        return False, self
 
 
 _typevariable_counter = count()
 
 
 class TypeVariable:
-    __slots__ = ("idx",)
+    __slots__ = ("_cached_hash", "idx", "name")
 
-    def __init__(self, idx: Optional[int] = None):
+    def __init__(self, idx: int | None = None, name: str | None = None):
         if idx is None:
             self.idx: int = next(_typevariable_counter)
         else:
             self.idx: int = idx
+        self.name = name
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
-        varname = mapping.get(self, None)
+        self._cached_hash = hash((TypeVariable, self.name if self.name else self.idx))
+
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
+        varname = mapping.get(self, self.name)
         if varname is None:
             return repr(self)
-        return f"{varname} ({repr(self)})"
+        return f"{varname} ({self!r})"
 
     def __eq__(self, other):
-        return type(other) is TypeVariable and other.idx == self.idx
+        if type(other) is not TypeVariable:
+            return False
+        if self.name or other.name:
+            return self.name == other.name
+        return self.idx == other.idx
+
+    def _hash(self, visited=None):  # pylint:disable=unused-argument
+        return self._cached_hash
 
     def __hash__(self):
-        return hash((TypeVariable, self.idx))
+        return self._cached_hash
 
     def __repr__(self):
-        return "tv_%02d" % self.idx
+        if self.name:
+            return f"{self.name}|tv_{self.idx:02d}"
+        return f"tv_{self.idx:02d}"
 
 
 class DerivedTypeVariable(TypeVariable):
-    __slots__ = (
-        "type_var",
-        "label",
-    )
+    __slots__ = ("labels", "type_var")
 
-    def __init__(self, type_var, label, idx=None):
+    type_var: TypeVariable
+    labels: tuple[BaseLabel, ...]
+
+    def __init__(
+        self,
+        type_var: TypeVariable | DerivedTypeVariable,
+        label: BaseLabel | None,
+        labels: Iterable[BaseLabel] | None = None,
+        idx=None,
+    ):
         super().__init__(idx=idx)
-        self.type_var = type_var
-        self.label = label
+        if isinstance(type_var, DerivedTypeVariable):
+            existing_labels = type_var.labels
+            self.type_var = type_var.type_var
+            assert not isinstance(self.type_var, DerivedTypeVariable)
+        else:
+            existing_labels = ()
+            self.type_var = type_var
 
-    def pp_str(self, mapping: Dict["TypeVariable", Any]) -> str:
-        return f"{self.type_var.pp_str(mapping)}.{self.label}"
+        if label is not None and labels:
+            raise TypeError("You cannot specify both label and labels at the same time")
+
+        if label is not None:
+            self.labels = (*existing_labels, label)
+        elif labels is not None:
+            self.labels = existing_labels + tuple(labels)
+        else:
+            self.labels = existing_labels
+
+        if not self.labels:
+            raise ValueError("A DerivedTypeVariable must have at least one label")
+
+        self._cached_hash = hash((DerivedTypeVariable, self.type_var, self.labels))
+
+    def one_label(self) -> BaseLabel | None:
+        return self.labels[0] if len(self.labels) == 1 else None
+
+    def path(self) -> tuple[BaseLabel, ...]:
+        return self.labels
+
+    def longest_prefix(self) -> TypeType | None:
+        if not self.labels:
+            return None
+        if len(self.labels) == 1:
+            return self.type_var
+        return DerivedTypeVariable(self.type_var, None, labels=self.labels[:-1])
+
+    def pp_str(self, mapping: dict[TypeVariable, Any]) -> str:
+        return ".".join([self.type_var.pp_str(mapping)] + [repr(lbl) for lbl in self.labels])
 
     def __eq__(self, other):
-        return isinstance(other, DerivedTypeVariable) and self.type_var == other.type_var and self.label == other.label
+        return (
+            isinstance(other, DerivedTypeVariable) and self.type_var == other.type_var and self.labels == other.labels
+        )
+
+    def _hash(self, visited=None):
+        return self._cached_hash
 
     def __hash__(self):
-        return hash((DerivedTypeVariable, self.type_var, self.label))
+        return self._cached_hash
 
     def __repr__(self):
-        return f"{self.type_var!r}.{self.label!r}"
+        return ".".join([repr(self.type_var)] + [repr(lbl) for lbl in self.labels])
 
     def replace(self, replacements):
         typevar = None
@@ -334,31 +392,27 @@ class DerivedTypeVariable(TypeVariable):
 
         if typevar is not None:
             # replacement has happened
-            return True, DerivedTypeVariable(typevar, self.label, idx=self.idx)
-        else:
-            return False, self
+            return True, DerivedTypeVariable(typevar, None, labels=self.labels, idx=self.idx)
+        return False, self
 
 
 class TypeVariables:
-    __slots__ = ("_typevars",)
+    __slots__ = (
+        "_last_typevars",
+        "_typevar2var",
+        "_typevars",
+    )
 
     def __init__(self):
-        self._typevars: Dict["SimVariable", TypeVariable] = ChainMapCOW(collapse_threshold=25)
-
-    def merge(self, tvs):
-        merged = TypeVariables()
-
-        # TODO: Replace this with a real lattice-based merging
-        merged._typevars = self._typevars.copy()
-        if tvs._typevars:
-            merged._typevars = merged._typevars.clean()
-            merged._typevars.update(tvs._typevars)
-
-        return merged
+        self._typevars: dict[SimVariable, set[TypeVariable]] = {}
+        self._typevar2var: dict[TypeVariable, SimVariable] = {}
+        self._last_typevars: dict[SimVariable, TypeVariable] = {}
 
     def copy(self):
         copied = TypeVariables()
-        copied._typevars = self._typevars.copy()
+        for var, typevars in self._typevars.items():
+            copied._typevars[var] = typevars.copy()
+        copied._last_typevars = self._last_typevars.copy()
         return copied
 
     def __repr__(self):
@@ -366,30 +420,29 @@ class TypeVariables:
         #    sum(len(v) for v in self._typevars.items()),
         #    len(self._typevars),
         # )
-        return "{TypeVars: %d items}" % len(self._typevars)
+        return f"{{TypeVars: {len(self._typevars)} items}}"
 
-    def add_type_variable(self, var: "SimVariable", codeloc, typevar: TypeVariable):  # pylint:disable=unused-argument
-        # if var not in self._typevars:
-        #    self._typevars[var] = { }
-
-        # assert codeloc not in self._typevars[var]
-        # self._typevars[var][codeloc] = typevar
-        self._typevars = self._typevars.clean()
-        self._typevars[var] = typevar
-
-    def get_type_variable(self, var, codeloc):  # pylint:disable=unused-argument
-        return self._typevars[var]  # [codeloc]
-
-    def has_type_variable_for(self, var: "SimVariable", codeloc):  # pylint:disable=unused-argument
+    def add_type_variable(self, var: SimVariable, typevar: TypeVariable, latest: bool = True):
         if var not in self._typevars:
-            return False
-        return True
-        # if codeloc not in self._typevars[var]:
-        #     return False
-        # return True
+            self._typevars[var] = set()
+        elif typevar in self._typevars[var]:
+            return
+        self._typevars[var].add(typevar)
+        if latest:
+            self._last_typevars[var] = typevar
+        self._typevar2var[typevar] = var
+
+    def get_type_variable(self, var):  # pylint:disable=unused-argument
+        return self._last_typevars[var]
+
+    def has_type_variable_for(self, var: SimVariable):  # pylint:disable=unused-argument
+        return var in self._typevars
+
+    def typevar_to_variable(self, typevar: TypeVariable) -> SimVariable | None:
+        return self._typevar2var.get(typevar, None)
 
     def __getitem__(self, var):
-        return self._typevars[var]
+        return self._last_typevars[var]
 
     def __contains__(self, var):
         return var in self._typevars
@@ -401,13 +454,20 @@ class TypeVariables:
 
 
 class BaseLabel:
-    __slots__ = ()
+    __slots__ = ("_cached_hash",)
+
+    def __init__(self):
+        self._cached_hash = hash((type(self), *tuple(getattr(self, k) for k in self.__slots__ if k != "_cached_hash")))
 
     def __eq__(self, other):
-        return type(self) is type(other) and hash(self) == hash(other)
+        return type(self) is type(other) and self._cached_hash == other._cached_hash
 
     def __hash__(self):
-        return hash(tuple(getattr(self, k) for k in self.__slots__))
+        return self._cached_hash
+
+    @property
+    def variance(self) -> Variance:
+        return Variance.COVARIANT
 
 
 class FuncIn(BaseLabel):
@@ -415,9 +475,10 @@ class FuncIn(BaseLabel):
 
     def __init__(self, loc):
         self.loc = loc
+        super().__init__()
 
     def __repr__(self):
-        return "in<%s>" % self.loc
+        return f"in<{self.loc}>"
 
 
 class FuncOut(BaseLabel):
@@ -425,9 +486,10 @@ class FuncOut(BaseLabel):
 
     def __init__(self, loc):
         self.loc = loc
+        super().__init__()
 
     def __repr__(self):
-        return "out<%s>" % self.loc
+        return f"out<{self.loc}>"
 
 
 class Load(BaseLabel):
@@ -443,15 +505,20 @@ class Store(BaseLabel):
     def __repr__(self):
         return "store"
 
+    @property
+    def variance(self) -> Variance:
+        return Variance.CONTRAVARIANT
+
 
 class AddN(BaseLabel):
     __slots__ = ("n",)
 
     def __init__(self, n):
         self.n = n
+        super().__init__()
 
     def __repr__(self):
-        return "+%d" % self.n
+        return f"+{self.n}"
 
 
 class SubN(BaseLabel):
@@ -459,9 +526,10 @@ class SubN(BaseLabel):
 
     def __init__(self, n):
         self.n = n
+        super().__init__()
 
     def __repr__(self):
-        return "-%d" % self.n
+        return f"-{self.n}"
 
 
 class ConvertTo(BaseLabel):
@@ -469,20 +537,22 @@ class ConvertTo(BaseLabel):
 
     def __init__(self, to_bits):
         self.to_bits = to_bits
+        super().__init__()
 
     def __repr__(self):
-        return "conv(%d)" % self.to_bits
+        return f"conv({self.to_bits})"
 
 
 class ReinterpretAs(BaseLabel):
     __slots__ = (
-        "to_type",
         "to_bits",
+        "to_type",
     )
 
     def __init__(self, to_type, to_bits):
         self.to_type = to_type
         self.to_bits = to_bits
+        super().__init__()
 
     def __repr__(self):
         return f"reinterpret({self.to_type}{self.to_bits})"
@@ -497,11 +567,56 @@ class HasField(BaseLabel):
     def __init__(self, bits, offset):
         self.bits = bits
         self.offset = offset
+        super().__init__()
 
     def __repr__(self):
-        return "<%d>@%d" % (self.bits, self.offset)
+        if self.bits == MAX_POINTSTO_BITS:
+            return f"<MAX_POINTSTO_BITS>@{self.offset}"
+        return f"<{self.bits} bits>@{self.offset}"
 
 
 class IsArray(BaseLabel):
     def __repr__(self):
         return "is_array"
+
+
+def new_dtv(
+    type_var: TypeVariable,
+    *,
+    label: BaseLabel | None = None,
+    labels: Sequence[BaseLabel] | None = None,
+) -> TypeVariable | DerivedTypeVariable:
+    """
+    Create a new DerivedTypeVariable with the given type variable (or DerivedTypeVariable) and labels.
+    """
+
+    if label is None and labels is None:
+        raise ValueError("Either label or labels must be specified")
+    new_labels = (label,) if label is not None else tuple(labels)  # type: ignore[reportArgumentType]
+    if isinstance(type_var, DerivedTypeVariable):
+        base_typevar = type_var.type_var
+        new_labels = type_var.labels + new_labels
+    else:
+        base_typevar = type_var
+
+    # condense the last N labels if they are AddN and SubN
+    off = 1
+    while off <= len(new_labels) and isinstance(new_labels[-off], (AddN, SubN)):
+        off += 1
+    if off <= len(new_labels) and not isinstance(new_labels[-off], (AddN, SubN)):
+        off -= 1
+
+    if off >= 2:
+        new_n = 0
+        for lbl in new_labels[-off:]:
+            if isinstance(lbl, AddN):
+                new_n += lbl.n
+            elif isinstance(lbl, SubN):
+                new_n -= lbl.n
+        new_labels = new_labels[:-off]
+        if new_n > 0:
+            new_labels += (AddN(new_n),)
+        elif new_n < 0:
+            new_labels += (SubN(-new_n),)
+
+    return DerivedTypeVariable(base_typevar, None, labels=new_labels) if new_labels else base_typevar
