@@ -1,5 +1,6 @@
 # pylint:disable=wrong-import-position,wrong-import-order
-from typing import Optional, List, Tuple, Union, DefaultDict, Set, Dict, TYPE_CHECKING
+from __future__ import annotations
+from typing import TYPE_CHECKING
 import logging
 from collections import defaultdict
 
@@ -7,20 +8,28 @@ import networkx
 
 import claripy
 import pyvex
-import ailment
+import angr.ailment as ailment
+from angr.ailment.expression import VirtualVariable
 
 import angr.errors
-from ...storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
-from ...block import Block
-from ...errors import AngrVariableRecoveryError, SimEngineError
-from ...knowledge_plugins import Function
-from ...sim_variable import SimStackVariable, SimRegisterVariable, SimVariable, SimMemoryVariable
-from ...engines.vex.claripy.irop import vexop_to_simop
+from angr import SIM_TYPE_COLLECTIONS
+from angr.analyses import AnalysesHub
+from angr.storage.memory_mixins.paged_memory.pages.multi_values import MultiValues
+from angr.block import Block
+from angr.errors import AngrVariableRecoveryError, SimEngineError
+from angr.knowledge_plugins import Function
+from angr.knowledge_plugins.key_definitions import atoms
+from angr.sim_variable import SimStackVariable, SimRegisterVariable, SimVariable, SimMemoryVariable
+from angr.engines.vex.claripy.irop import vexop_to_simop
 from angr.analyses import ForwardAnalysis, visitors
-from ..typehoon.typevars import Equivalence, TypeVariable, TypeVariables
+from angr.analyses.typehoon.typevars import Equivalence, TypeVariable, TypeVariables, Subtype, DerivedTypeVariable
+from angr.analyses.typehoon.typeconsts import Int, TypeConstant, BottomType, TopType
+from angr.analyses.typehoon.lifter import TypeLifter
 from .variable_recovery_base import VariableRecoveryBase, VariableRecoveryStateBase
 from .engine_vex import SimEngineVRVEX
 from .engine_ail import SimEngineVRAIL
+import contextlib
+
 
 if TYPE_CHECKING:
     from angr.analyses.typehoon.typevars import TypeConstraint
@@ -42,6 +51,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         analysis,
         arch,
         func,
+        project,
         stack_region=None,
         register_region=None,
         global_region=None,
@@ -50,7 +60,6 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         func_typevar=None,
         delayed_type_constraints=None,
         stack_offset_typevars=None,
-        project=None,
         ret_val_size=None,
     ):
         super().__init__(
@@ -58,6 +67,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             analysis,
             arch,
             func,
+            project,
             stack_region=stack_region,
             register_region=register_region,
             global_region=global_region,
@@ -66,16 +76,11 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             func_typevar=func_typevar,
             delayed_type_constraints=delayed_type_constraints,
             stack_offset_typevars=stack_offset_typevars,
-            project=project,
         )
         self.ret_val_size = ret_val_size
 
     def __repr__(self):
-        return "<VRAbstractState@%#x: %d register variables, %d stack variables>" % (
-            self.block_addr,
-            len(self.register_region),
-            len(self.stack_region),
-        )
+        return f"<VRAbstractState@{self.block_addr:#x}"
 
     def __eq__(self, other):
         if type(other) is not VariableRecoveryFastState:
@@ -83,7 +88,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         return self.stack_region == other.stack_region and self.register_region == other.register_region
 
     def copy(self):
-        state = VariableRecoveryFastState(
+        return VariableRecoveryFastState(
             self.block_addr,
             self._analysis,
             self.arch,
@@ -95,16 +100,14 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             type_constraints=self.type_constraints,
             func_typevar=self.func_typevar,
             delayed_type_constraints=self.delayed_type_constraints,
-            stack_offset_typevars=dict(self.stack_offset_typevars),
+            stack_offset_typevars=self.stack_offset_typevars,
             project=self.project,
             ret_val_size=self.ret_val_size,
         )
 
-        return state
-
     def merge(
-        self, others: Tuple["VariableRecoveryFastState"], successor=None
-    ) -> Tuple["VariableRecoveryFastState", bool]:
+        self, others: tuple[VariableRecoveryFastState, ...], successor=None
+    ) -> tuple[VariableRecoveryFastState, bool]:
         """
         Merge two abstract states.
 
@@ -138,10 +141,10 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
         # add subtype constraints for all replacements
         for v0, v1 in self.phi_variables.items():
             # v0 will be replaced by v1
-            if not typevars.has_type_variable_for(v1, None):
-                typevars.add_type_variable(v1, None, TypeVariable())
-            if not typevars.has_type_variable_for(v0, None):
-                typevars.add_type_variable(v0, None, TypeVariable())
+            if not typevars.has_type_variable_for(v1):
+                typevars.add_type_variable(v1, TypeVariable())
+            if not typevars.has_type_variable_for(v0):
+                typevars.add_type_variable(v0, TypeVariable())
             # Assuming v2 = phi(v0, v1), then we know that v0_typevar == v1_typevar == v2_typevar
             # However, it's possible that neither v0 nor v1 will ever be used in future blocks, which not only makes
             # this phi function useless, but also leads to the incorrect assumption that v1_typevar == v2_typevar.
@@ -149,7 +152,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             # when v1 (the new variable that will end up in the state) is ever used in the future.
 
             # create an equivalence relationship
-            equivalence = Equivalence(typevars.get_type_variable(v1, None), typevars.get_type_variable(v0, None))
+            equivalence = Equivalence(typevars.get_type_variable(v1), typevars.get_type_variable(v0))
             delayed_typeconstraints[v1].add(equivalence)
 
         stack_offset_typevars = {}
@@ -158,7 +161,7 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
             all_stack_addr_typevar_offsets.update(other.stack_offset_typevars)
         for offset in all_stack_addr_typevar_offsets:
             all_typevars = set()
-            for state in (self,) + others:
+            for state in (self, *others):
                 typevar = state.stack_offset_typevars.get(offset, None)
                 if typevar is not None:
                     all_typevars.add(typevar)
@@ -173,10 +176,9 @@ class VariableRecoveryFastState(VariableRecoveryStateBase):
 
         ret_val_size = self.ret_val_size
         for o in others:
-            if o.ret_val_size is not None:
-                if ret_val_size is None or o.ret_val_size > ret_val_size:
-                    ret_val_size = o.ret_val_size
-                    merge_occurred = True
+            if o.ret_val_size is not None and (ret_val_size is None or o.ret_val_size > ret_val_size):
+                ret_val_size = o.ret_val_size
+                merge_occurred = True
 
         # clean up
         self.phi_variables = {}
@@ -233,14 +235,18 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
 
     def __init__(
         self,
-        func: Union[Function, str, int],
-        func_graph: Optional[networkx.DiGraph] = None,
+        func: Function | str | int,
+        func_graph: networkx.DiGraph | None = None,
+        entry_node_addr: int | tuple[int, int | None] | None = None,
         max_iterations: int = 2,
         low_priority=False,
         track_sp=True,
-        func_args: Optional[List[SimVariable]] = None,
+        func_args: list[SimVariable] | None = None,
         store_live_variables=False,
         unify_variables=True,
+        func_arg_vvars: dict[int, tuple[VirtualVariable, SimVariable]] | None = None,
+        vvar_to_vvar: dict[int, int] | None = None,
+        type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]] | None = None,
     ):
         if not isinstance(func, Function):
             func = self.kb.functions[func]
@@ -248,18 +254,24 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         call_info = defaultdict(list)
         for node_from, node_to, data in func_graph_with_calls.edges(data=True):
             if data.get("type", None) == "call" or data.get("outside", False):
-                try:
+                with contextlib.suppress(KeyError):
                     call_info[node_from.addr].append(self.kb.functions.get_by_addr(node_to.addr))
-                except KeyError:
-                    pass
 
         function_graph_visitor = visitors.FunctionGraphVisitor(func, graph=func_graph)
 
         # Make sure the function is not empty
-        if not func.block_addrs_set or func.startpoint is None:
-            raise AngrVariableRecoveryError("Function %s is empty." % repr(func))
+        if (not func.block_addrs_set or func.startpoint is None) and not func_graph:
+            raise AngrVariableRecoveryError(f"Function {func!r} is empty.")
 
-        VariableRecoveryBase.__init__(self, func, max_iterations, store_live_variables)
+        VariableRecoveryBase.__init__(
+            self,
+            func,
+            max_iterations,
+            store_live_variables,
+            vvar_to_vvar=vvar_to_vvar,
+            func_graph=func_graph_with_calls,
+            entry_node_addr=entry_node_addr,
+        )
         ForwardAnalysis.__init__(
             self, order_jobs=True, allow_merging=True, allow_widening=False, graph_visitor=function_graph_visitor
         )
@@ -268,27 +280,40 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
         self._job_ctr = 0
         self._track_sp = track_sp and self.project.arch.sp_offset is not None
         self._func_args = func_args
+        self._func_arg_vvars = func_arg_vvars
         self._unify_variables = unify_variables
 
-        self._ail_engine = SimEngineVRAIL(self.project, self.kb, call_info=call_info)
-        self._vex_engine = SimEngineVRVEX(self.project, self.kb, call_info=call_info)
+        # handle type hints
+        self.vvar_type_hints = {}
+        if type_hints:
+            self._parse_type_hints(type_hints)
+
+        self._ail_engine: SimEngineVRAIL = SimEngineVRAIL(
+            self.project,
+            self.kb,
+            call_info=call_info,
+            vvar_to_vvar=self.vvar_to_vvar,
+            vvar_type_hints=self.vvar_type_hints,
+        )
+        self._vex_engine: SimEngineVRVEX = SimEngineVRVEX(self.project, self.kb, call_info=call_info)
 
         self._node_iterations = defaultdict(int)
 
         self._node_to_cc = {}
-        self.var_to_typevars: DefaultDict[SimVariable, Set[TypeVariable]] = defaultdict(set)
+        self.var_to_typevars: defaultdict[SimVariable, set[TypeVariable]] = defaultdict(set)
         self.typevars = None
-        self.type_constraints: Optional[Dict["TypeVariable", Set["TypeConstraint"]]] = None
+        self.type_constraints: dict[TypeVariable, set[TypeConstraint]] | None = None
         self.func_typevar = TypeVariable(name=func.name)
         self.delayed_type_constraints = None
         self.ret_val_size = None
+        self.stack_offset_typevars: dict[int, TypeVariable] = {}
 
         self._analyze()
 
         # cleanup (for cpython pickle)
         self.downsize()
-        self._ail_engine = None
-        self._vex_engine = None
+        del self._ail_engine
+        del self._vex_engine
 
     #
     # Main analysis routines
@@ -329,6 +354,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             type_constraints=self.type_constraints,
             func_typevar=self.func_typevar,
             delayed_type_constraints=self.delayed_type_constraints,
+            stack_offset_typevars=self.stack_offset_typevars,
         )
         initial_sp = state.stack_address(self.project.arch.bytes if self.project.arch.call_pushes_ret else 0)
         if self.project.arch.sp_offset is not None:
@@ -377,7 +403,27 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             except angr.errors.SimMemoryMissingError:
                 state.register_region.store(t9_offset, claripy.BVV(node.addr, t9_size * 8))
 
-        if self._func_args:
+        if self._func_arg_vvars:
+            for arg_vvar, arg in self._func_arg_vvars.values():
+                if isinstance(arg, SimRegisterVariable):
+                    v = claripy.BVS("reg_arg", arg.bits)
+                    v = state.annotate_with_variables(v, [(0, arg)])
+                    arg_vvar_id = arg_vvar.varid
+                    if self.vvar_to_vvar:
+                        arg_vvar_id = self.vvar_to_vvar.get(arg_vvar_id, arg_vvar_id)
+                    self._ail_engine.vvar_region[arg_vvar_id] = v
+                    internal_manager.add_variable("register", arg.reg, arg)
+                elif isinstance(arg, SimStackVariable):
+                    v = claripy.BVS("stack_arg", arg.bits)
+                    v = state.annotate_with_variables(v, [(0, arg)])
+                    arg_vvar_id = arg_vvar.varid
+                    if self.vvar_to_vvar:
+                        arg_vvar_id = self.vvar_to_vvar.get(arg_vvar_id, arg_vvar_id)
+                    self._ail_engine.vvar_region[arg_vvar_id] = v
+                    internal_manager.add_variable("stack", arg.offset, arg)
+                else:
+                    raise TypeError(f"Unsupported function argument type {type(arg)}")
+        elif self._func_args:
             for arg in self._func_args:
                 if isinstance(arg, SimRegisterVariable):
                     v = claripy.BVS("reg_arg", arg.bits)
@@ -394,7 +440,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
                     )
                     internal_manager.add_variable("stack", arg.offset, arg)
                 else:
-                    raise TypeError("Unsupported function argument type %s." % type(arg))
+                    raise TypeError(f"Unsupported function argument type {type(arg)}.")
 
         return state
 
@@ -420,20 +466,10 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             block = self.project.factory.block(node.addr, node.size, opt_level=1, cross_insn_opt=False)
             block_key = node.addr
 
-        # if node.addr in self._instates:
-        #     prev_state: VariableRecoveryFastState = self._instates[node.addr]
-        #     if input_state == prev_state:
-        #         l.debug('Skip node %#x as we have reached a fixed-point', node.addr)
-        #         return False, input_state
-        #     else:
-        #         l.debug('Merging input state of node %#x with the previous state.', node.addr)
-        #         input_state, _ = prev_state.merge((input_state,), successor=node.addr)
-
         state = state.copy()
         state.block_addr = node.addr
         if isinstance(node, ailment.Block):
             state.block_idx = node.idx
-        # self._instates[node.addr] = state
 
         if self._node_iterations[block_key] >= self._max_iterations:
             l.debug("Skip node %#x as we have iterated %d times on it.", node.addr, self._node_iterations[node.addr])
@@ -443,9 +479,8 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
 
         self._node_iterations[block_key] += 1
 
-        if state.ret_val_size is not None:
-            if self.ret_val_size is None or self.ret_val_size < state.ret_val_size:
-                self.ret_val_size = state.ret_val_size
+        if state.ret_val_size is not None and (self.ret_val_size is None or self.ret_val_size < state.ret_val_size):
+            self.ret_val_size = state.ret_val_size
 
         state.downsize()
         self._outstates[block_key] = state
@@ -473,16 +508,38 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             self.variable_manager[self.function.addr].unify_variables()
 
         # fill in var_to_typevars
+        assert self.typevars is not None
         for var, typevar_set in self.typevars._typevars.items():
             self.var_to_typevars[var] = typevar_set
 
         # unify type variables for global variables
+        assert self.type_constraints is not None
         for var, typevars in self.var_to_typevars.items():
             if len(typevars) > 1 and isinstance(var, SimMemoryVariable) and not isinstance(var, SimStackVariable):
-                sorted_typevars = list(sorted(typevars, key=lambda x: str(x)))  # pylint:disable=unnecessary-lambda
+                sorted_typevars = sorted(typevars, key=lambda x: str(x))  # pylint:disable=unnecessary-lambda
                 for tv in sorted_typevars[1:]:
                     self.type_constraints[self.func_typevar].add(Equivalence(sorted_typevars[0], tv))
 
+        # remove default constraints with size conflicts
+        for func_var in self.type_constraints:
+            var_to_subtyping: dict[TypeVariable, list[Subtype]] = defaultdict(list)
+            for constraint in self.type_constraints[func_var]:
+                if isinstance(constraint, Subtype) and isinstance(constraint.sub_type, TypeVariable):
+                    var_to_subtyping[constraint.sub_type].append(constraint)
+
+            for constraints in var_to_subtyping.values():
+                if len(constraints) <= 1:
+                    continue
+                default_subtyping_constraints = set()
+                has_nondefault_subtyping_constraints = False
+                for constraint in constraints:
+                    if isinstance(constraint.super_type, Int):
+                        default_subtyping_constraints.add(constraint)
+                    elif isinstance(constraint.super_type, DerivedTypeVariable) and constraint.super_type.labels:
+                        has_nondefault_subtyping_constraints = True
+                if has_nondefault_subtyping_constraints:
+                    self.type_constraints[func_var].difference_update(default_subtyping_constraints)
+
         self.variable_manager[self.function.addr].ret_val_size = self.ret_val_size
 
         self.delayed_type_constraints = None
@@ -503,7 +560,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             256: pyvex.const.V256,
         }
         if size not in mapping:
-            raise TypeError("Unsupported size %d." % size)
+            raise TypeError(f"Unsupported size {size}.")
         return mapping.get(size)(value)
 
     def _peephole_optimize(self, block: Block):
@@ -522,20 +579,21 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
                         tmp0 = stmt0.tmp
                         tmp1 = stmt1.tmp
                         stmt2 = block.vex.statements[i + 2]
-                        if isinstance(stmt2, pyvex.IRStmt.WrTmp) and isinstance(stmt2.data, pyvex.IRExpr.Binop):
-                            if (
-                                isinstance(stmt2.data.args[0], pyvex.IRExpr.RdTmp)
-                                and isinstance(stmt2.data.args[1], pyvex.IRExpr.RdTmp)
-                                and {stmt2.data.args[0].tmp, stmt2.data.args[1].tmp} == {tmp0, tmp1}
-                                and vexop_to_simop(stmt2.data.op)._generic_name == "Xor"
-                            ):
-                                # found it!
-                                # make a copy so we don't trash the cached VEX IRSB
-                                block._vex = block.vex.copy()
-                                block.vex.statements[i] = pyvex.IRStmt.NoOp()
-                                block.vex.statements[i + 1] = pyvex.IRStmt.NoOp()
-                                zero = pyvex.IRExpr.Const(self._get_irconst(0, block.vex.tyenv.sizeof(tmp0)))
-                                block.vex.statements[i + 2] = pyvex.IRStmt.Put(zero, reg_offset)
+                        if (
+                            isinstance(stmt2, pyvex.IRStmt.WrTmp)
+                            and isinstance(stmt2.data, pyvex.IRExpr.Binop)
+                            and isinstance(stmt2.data.args[0], pyvex.IRExpr.RdTmp)
+                            and isinstance(stmt2.data.args[1], pyvex.IRExpr.RdTmp)
+                            and {stmt2.data.args[0].tmp, stmt2.data.args[1].tmp} == {tmp0, tmp1}
+                            and vexop_to_simop(stmt2.data.op)._generic_name == "Xor"
+                        ):
+                            # found it!
+                            # make a copy so we don't trash the cached VEX IRSB
+                            block._vex = block.vex.copy()
+                            block.vex.statements[i] = pyvex.IRStmt.NoOp()
+                            block.vex.statements[i + 1] = pyvex.IRStmt.NoOp()
+                            zero = pyvex.IRExpr.Const(self._get_irconst(0, block.vex.tyenv.sizeof(tmp0)))
+                            block.vex.statements[i + 2] = pyvex.IRStmt.Put(zero, reg_offset)
             i = next_i
         return block
 
@@ -561,7 +619,7 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
             block = self._peephole_optimize(block)
 
         processor = self._ail_engine if isinstance(block, ailment.Block) else self._vex_engine
-        processor.process(state, block=block, fail_fast=self._fail_fast)
+        processor.process(state, block=block, fail_fast=self._fail_fast)  # type: ignore
 
         if self._track_sp and block.addr in self._node_to_cc:
             # readjusting sp at the end for blocks that end in a call
@@ -573,17 +631,31 @@ class VariableRecoveryFast(ForwardAnalysis, VariableRecoveryBase):  # pylint:dis
 
             adjusted = False
 
-            if not adjusted:
-                # make a guess
-                # of course, this will fail miserably if the function called is not cdecl
-                if self.project.arch.call_pushes_ret:
-                    sp_v += self.project.arch.bytes
-                    adjusted = True
+            # make a guess
+            # of course, this will fail miserably if the function called is not cdecl
+            if not adjusted and self.project.arch.call_pushes_ret:
+                sp_v += self.project.arch.bytes
+                adjusted = True
 
             if adjusted:
                 state.register_region.store(self.project.arch.sp_offset, sp_v)
 
+    def _parse_type_hints(self, type_hints: list[tuple[atoms.VirtualVariable | atoms.MemoryLocation, str]]) -> None:
+        self.vvar_type_hints = {}
+        for loc, type_hint_str in type_hints:
+            if isinstance(loc, atoms.VirtualVariable):
+                type_hint = self._parse_type_hint(type_hint_str)
+                if type_hint is not None:
+                    self.vvar_type_hints[loc.varid] = type_hint
+            # TODO: Handle other types of locations
+
+    def _parse_type_hint(self, type_hint_str: str) -> TypeConstant | None:
+        ty = SIM_TYPE_COLLECTIONS["cpp::std"].get(type_hint_str)
+        if ty is None:
+            return None
+        ty = ty.with_arch(self.project.arch)
+        lifted = TypeLifter(self.project.arch.bits).lift(ty)
+        return None if isinstance(lifted, (BottomType, TopType)) else lifted
 
-from angr.analyses import AnalysesHub
 
 AnalysesHub.register_default("VariableRecoveryFast", VariableRecoveryFast)
