@@ -1,13 +1,15 @@
-from typing import Dict, Optional
+from __future__ import annotations
 import logging
 
 import claripy
 from archinfo.arch_arm import is_arm_arch
-from angr.state_plugins.sim_action_object import _raw_ast, SimActionObject
+
 from angr import errors
+from angr.errors import SimError, SimCCallError
+from angr.sim_options import USE_SIMPLIFIED_CCALLS
+from angr.state_plugins.sim_action_object import _raw_ast, SimActionObject
 
 l = logging.getLogger(name=__name__)
-# l.setLevel(logging.DEBUG)
 
 # pylint: disable=R0911
 # pylint: disable=W0613
@@ -21,10 +23,7 @@ l = logging.getLogger(name=__name__)
 
 # There might be a better way of doing this
 def calc_paritybit(p, msb=7, lsb=0):
-    if len(p) > msb:
-        p_part = p[msb:lsb]
-    else:
-        p_part = p
+    p_part = p[msb:lsb] if len(p) > msb else p
 
     b = claripy.BVV(1, 1)
     for i in range(p_part.size()):
@@ -47,21 +46,49 @@ def boolean_extend(O, a, b, size):
 def op_concretize(op):
     if type(op) is int:
         return op
-    op_e = op.ite_excavated
+    op_e = claripy.excavate_ite(op)
     if op_e.op == "If":
         cases = list(claripy.reverse_ite_cases(op_e))
         if all(c.op == "BVV" for _, c in cases):
             raise CCallMultivaluedException(cases, op)
+    if op.op == "__or__":
+        # I've seen cases like this in AARCH64:
+        #    CCMP            W9, W8, #4, NE
+        #
+        #    00 | ------ IMark(0x0, 4, 0) ------
+        #    01 | t12 = GET:I64(cc_op)
+        #    02 | t11 = Or64(t12,0x0000000000000010)
+        #    03 | t13 = GET:I64(cc_dep1)
+        #    04 | t14 = GET:I64(cc_dep2)
+        #    05 | t15 = GET:I64(cc_ndep)
+        #    06 | t16 = arm64g_calculate_condition(t11,t13,t14,t15):Ity_I64
+        #    07 | t10 = 64to1(t16)
+        #    08 | t18 = GET:I64(x9)
+        #    09 | t17 = 64to32(t18)
+        #    10 | t20 = GET:I64(x8)
+        #    11 | t19 = 64to32(t20)
+        #    12 | t21 = 32Uto64(t17)
+        #    13 | t22 = 32Uto64(t19)
+        #    14 | t23 = ITE(t10,0x0000000000000003,0x0000000000000000)
+        #    15 | t24 = ITE(t10,t21,0x0000000040000000)
+        #    16 | t25 = ITE(t10,t22,0x0000000000000000)
+        #    17 | PUT(cc_op) = t23
+        #    18 | PUT(cc_dep1) = t24
+        #    19 | PUT(cc_dep2) = t25
+        #    20 | PUT(cc_ndep) = 0x0000000000000000
+        if op.args[0].op == "BVV":
+            return op.args[0].args[0]
+        if op.args[1].op == "BVV":
+            return op.args[1].args[0]
     if op.op != "BVV":
         raise SimError(
-            "Hit a symbolic conditional operation (need If or BVV, got %s). " % op.op
-            + "Something has gone wildly wrong."
+            f"Hit a symbolic conditional operation (need If or BVV, got {op.op}). " + "Something has gone wildly wrong."
         )
     return op.args[0]
 
 
 def strip_simaction(val):
-    if type(val) == SimActionObject:
+    if isinstance(val, SimActionObject):
         return _raw_ast(val)
     return val
 
@@ -74,7 +101,7 @@ class CCallMultivaluedException(Exception):
 ### x86* data ###
 ##################
 
-data: Dict[str, Dict[str, Dict[str, Optional[int]]]] = {
+data: dict[str, dict[str, dict[str, int | None]]] = {
     "AMD64": {
         "CondTypes": {},
         "CondBitOffsets": {},
@@ -483,14 +510,12 @@ def pc_actions_INC(state, nbits, res, _, cc_ndep, platform=None):
 
 
 def pc_actions_SHL(state, nbits, remaining, shifted, cc_ndep, platform=None):
-    cf = ((remaining >> (nbits - 1)) & data[platform]["CondBitMasks"]["G_CC_MASK_C"])[
-        data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"]
-    ]
+    cf = shifted[nbits - 1]
     pf = calc_paritybit(remaining[7:0])
     af = claripy.BVV(0, 1)
     zf = calc_zerobit(remaining)
     sf = remaining[nbits - 1]
-    of = (remaining[0] ^ shifted[0])[0]
+    of = (sf ^ cf)[0]
     return pc_make_rdata(data[platform]["size"], cf, pf, af, zf, sf, of, platform=platform)
 
 
@@ -500,7 +525,7 @@ def pc_actions_SHR(state, nbits, remaining, shifted, cc_ndep, platform=None):
     af = claripy.BVV(0, 1)
     zf = calc_zerobit(remaining)
     sf = remaining[nbits - 1]
-    of = (remaining[0] ^ shifted[0])[0]
+    of = (remaining[nbits - 1] ^ shifted[nbits - 1])[0]
     return pc_make_rdata(data[platform]["size"], cf, pf, af, zf, sf, of, platform=platform)
 
 
@@ -543,9 +568,9 @@ def pc_actions_UMULQ(*args, **kwargs):
 
 
 def pc_actions_SMUL(state, nbits, cc_dep1, cc_dep2, cc_ndep, platform=None):
-    lo = (cc_dep1 * cc_dep2)[nbits - 1 : 0]
-    rr = lo
-    hi = (rr >> nbits)[nbits - 1 : 0]
+    ir = cc_dep1.sign_extend(nbits) * cc_dep2.sign_extend(nbits)
+    hi = ir[2 * nbits - 1 : nbits]
+    lo = ir[nbits - 1 : 0]
     cf = claripy.If(hi != (lo >> (nbits - 1)), claripy.BVV(1, 1), claripy.BVV(0, 1))
     zf = calc_zerobit(lo)
     pf = calc_paritybit(lo)
@@ -668,8 +693,7 @@ def pc_calculate_rdata_all(state, cc_op, cc_dep1, cc_dep2, cc_ndep, platform=Non
     rdata_all = pc_calculate_rdata_all_WRK(state, cc_op, cc_dep1, cc_dep2, cc_ndep, platform=platform)
     if isinstance(rdata_all, tuple):
         return pc_make_rdata_if_necessary(data[platform]["size"], *rdata_all, platform=platform)
-    else:
-        return rdata_all
+    return rdata_all
 
 
 # This function takes a condition that is being checked (ie, zero bit), and basically
@@ -730,59 +754,58 @@ def pc_calculate_condition(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep, platfo
             raise SimCCallError("Unrecognized condition in pc_calculate_condition. Panic.")
 
         return claripy.Concat(claripy.BVV(0, data[platform]["size"] - 1), r)
-    else:
-        rdata = rdata_all
-        v = op_concretize(cond)
-        inv = v & 1
-        l.debug("inv: %d", inv)
+    rdata = rdata_all
+    v = op_concretize(cond)
+    inv = v & 1
+    l.debug("inv: %d", inv)
 
-        # THIS IS A FUCKING HACK
-        if v == 0xE:
-            # jle
-            pass
-        if v in [data[platform]["CondTypes"]["CondO"], data[platform]["CondTypes"]["CondNO"]]:
-            l.debug("CondO")
-            of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
-            return 1 & (inv ^ of)
+    # THIS IS A FUCKING HACK
+    if v == 0xE:
+        # jle
+        pass
+    if v in [data[platform]["CondTypes"]["CondO"], data[platform]["CondTypes"]["CondNO"]]:
+        l.debug("CondO")
+        of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
+        return 1 & (inv ^ of)
 
-        if v in [data[platform]["CondTypes"]["CondZ"], data[platform]["CondTypes"]["CondNZ"]]:
-            l.debug("CondZ")
-            zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
-            return 1 & (inv ^ zf)
+    if v in [data[platform]["CondTypes"]["CondZ"], data[platform]["CondTypes"]["CondNZ"]]:
+        l.debug("CondZ")
+        zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
+        return 1 & (inv ^ zf)
 
-        if v in [data[platform]["CondTypes"]["CondB"], data[platform]["CondTypes"]["CondNB"]]:
-            l.debug("CondB")
-            cf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"])
-            return 1 & (inv ^ cf)
+    if v in [data[platform]["CondTypes"]["CondB"], data[platform]["CondTypes"]["CondNB"]]:
+        l.debug("CondB")
+        cf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"])
+        return 1 & (inv ^ cf)
 
-        if v in [data[platform]["CondTypes"]["CondBE"], data[platform]["CondTypes"]["CondNBE"]]:
-            l.debug("CondBE")
-            cf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"])
-            zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
-            return 1 & (inv ^ (cf | zf))
+    if v in [data[platform]["CondTypes"]["CondBE"], data[platform]["CondTypes"]["CondNBE"]]:
+        l.debug("CondBE")
+        cf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"])
+        zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
+        return 1 & (inv ^ (cf | zf))
 
-        if v in [data[platform]["CondTypes"]["CondS"], data[platform]["CondTypes"]["CondNS"]]:
-            l.debug("CondS")
-            sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
-            return 1 & (inv ^ sf)
+    if v in [data[platform]["CondTypes"]["CondS"], data[platform]["CondTypes"]["CondNS"]]:
+        l.debug("CondS")
+        sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
+        return 1 & (inv ^ sf)
 
-        if v in [data[platform]["CondTypes"]["CondP"], data[platform]["CondTypes"]["CondNP"]]:
-            l.debug("CondP")
-            pf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_P"])
-            return 1 & (inv ^ pf)
+    if v in [data[platform]["CondTypes"]["CondP"], data[platform]["CondTypes"]["CondNP"]]:
+        l.debug("CondP")
+        pf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_P"])
+        return 1 & (inv ^ pf)
 
-        if v in [data[platform]["CondTypes"]["CondL"], data[platform]["CondTypes"]["CondNL"]]:
-            l.debug("CondL")
-            sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
-            of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
-            return 1 & (inv ^ (sf ^ of))
+    if v in [data[platform]["CondTypes"]["CondL"], data[platform]["CondTypes"]["CondNL"]]:
+        l.debug("CondL")
+        sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
+        of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
+        return 1 & (inv ^ (sf ^ of))
 
-        if v in [data[platform]["CondTypes"]["CondLE"], data[platform]["CondTypes"]["CondNLE"]]:
-            l.debug("CondLE")
-            sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
-            of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
-            zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
-            return 1 & (inv ^ ((sf ^ of) | zf))
+    if v in [data[platform]["CondTypes"]["CondLE"], data[platform]["CondTypes"]["CondNLE"]]:
+        l.debug("CondLE")
+        sf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_S"])
+        of = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_O"])
+        zf = claripy.LShR(rdata, data[platform]["CondBitOffsets"]["G_CC_SHIFT_Z"])
+        return 1 & (inv ^ ((sf ^ of) | zf))
 
     l.error("Unsupported condition %d in in pc_calculate_condition", v)
     raise SimCCallError("Unrecognized condition in pc_calculate_condition")
@@ -937,8 +960,8 @@ def pc_calculate_condition_simple(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep,
     if funcname in globals():
         r = globals()[funcname](state, cc_dep1_nbits, cc_dep2_nbits, cc_ndep)
     else:
-        op_funcname = "pc_actions_op_%s" % op
-        cond_funcname = "pc_actions_cond_%s" % cond
+        op_funcname = f"pc_actions_op_{op}"
+        cond_funcname = f"pc_actions_cond_{cond}"
         if op_funcname in globals() and cond_funcname in globals():
             cc_expr = globals()[op_funcname](cc_dep1_nbits, cc_dep2_nbits, cc_ndep)
             r = globals()[cond_funcname](state, cc_expr)
@@ -959,7 +982,7 @@ def pc_calculate_rdata_c(state, cc_op, cc_dep1, cc_dep2, cc_ndep, platform=None)
 
     if cc_op == data[platform]["OpTypes"]["G_CC_OP_COPY"]:
         return claripy.LShR(cc_dep1, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"]) & 1  # TODO: actual constraints
-    elif cc_op in (
+    if cc_op in (
         data[platform]["OpTypes"]["G_CC_OP_LOGICQ"],
         data[platform]["OpTypes"]["G_CC_OP_LOGICL"],
         data[platform]["OpTypes"]["G_CC_OP_LOGICW"],
@@ -972,8 +995,7 @@ def pc_calculate_rdata_c(state, cc_op, cc_dep1, cc_dep2, cc_ndep, platform=None)
     if isinstance(rdata_all, tuple):
         cf, pf, af, zf, sf, of = rdata_all
         return claripy.Concat(claripy.BVV(0, data[platform]["size"] - 1), cf & 1)
-    else:
-        return claripy.LShR(rdata_all, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"]) & 1
+    return claripy.LShR(rdata_all, data[platform]["CondBitOffsets"]["G_CC_SHIFT_C"]) & 1
 
 
 def generic_rotate_with_carry(state, left, arg, rot_amt, carry_bit_in, sz):
@@ -991,10 +1013,7 @@ def generic_rotate_with_carry(state, left, arg, rot_amt, carry_bit_in, sz):
     if bits > len(rot_amt):
         raise SimError("Got a rotate instruction for data larger than the provided word size. Panic.")
 
-    if bits == len(rot_amt):
-        sized_amt = (rot_amt & (bits_in - 1)).zero_extend(1)
-    else:
-        sized_amt = (rot_amt & (bits_in - 1))[bits:0]
+    sized_amt = (rot_amt & bits_in - 1).zero_extend(1) if bits == len(rot_amt) else (rot_amt & bits_in - 1)[bits:0]
 
     assert len(sized_amt) == bits + 1
 
@@ -1101,8 +1120,7 @@ def amd64g_calculate_RCL(state, arg, rot_amt, eflags_in, sz):
             of << data["AMD64"]["CondBitOffsets"]["G_CC_SHIFT_O"]
         )
         return eflags_out
-    else:
-        return arg_out
+    return arg_out
 
 
 def amd64g_calculate_RCR(state, arg, rot_amt, eflags_in, sz):
@@ -1124,8 +1142,7 @@ def amd64g_calculate_RCR(state, arg, rot_amt, eflags_in, sz):
             of << data["AMD64"]["CondBitOffsets"]["G_CC_SHIFT_O"]
         )
         return eflags_out
-    else:
-        return arg_out
+    return arg_out
 
 
 def amd64g_calculate_mmx_pmaddwd(_state, xx, yy):
@@ -1201,8 +1218,7 @@ def x86g_calculate_RCR(state, arg, rot_amt, eflags_in, sz):
 def x86g_calculate_condition(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep):
     if USE_SIMPLIFIED_CCALLS in state.options:
         return pc_calculate_condition_simple(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep, platform="X86")
-    else:
-        return pc_calculate_condition(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep, platform="X86")
+    return pc_calculate_condition(state, cond, cc_op, cc_dep1, cc_dep2, cc_ndep, platform="X86")
 
 
 def x86g_calculate_eflags_all(state, cc_op, cc_dep1, cc_dep2, cc_ndep):
@@ -1294,7 +1310,7 @@ def x86g_calculate_daa_das_aaa_aas(state, flags_and_AX, opcode):
         r_C = claripy.If(condition, one, zero)
         r_O = r_S = r_Z = r_P = 0
 
-    result = (
+    return (
         ((r_O & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_O"]))
         | ((r_S & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_S"]))
         | ((r_Z & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_Z"]))
@@ -1304,7 +1320,6 @@ def x86g_calculate_daa_das_aaa_aas(state, flags_and_AX, opcode):
         | ((r_AH & 0xFF) << 8)
         | ((r_AL & 0xFF) << 0)
     )
-    return result
 
 
 def x86g_calculate_aad_aam(state, flags_and_AX, opcode):
@@ -1322,7 +1337,7 @@ def x86g_calculate_aad_aam(state, flags_and_AX, opcode):
         r_AL = ((r_AH * 10) + r_AL) & 0xFF
         r_AH = claripy.BVV(0, 32)
     else:
-        raise SimCCallError("Unknown opcode %#x in AAD/AAM ccall" % opcode)
+        raise SimCCallError(f"Unknown opcode {opcode:#x} in AAD/AAM ccall")
 
     r_O = claripy.BVV(0, 32)
     r_C = claripy.BVV(0, 32)
@@ -1331,7 +1346,7 @@ def x86g_calculate_aad_aam(state, flags_and_AX, opcode):
     r_Z = claripy.If(r_AL == 0, claripy.BVV(1, 32), claripy.BVV(0, 32))
     r_P = calc_paritybit(r_AL).zero_extend(31)
 
-    result = (
+    return (
         ((r_O & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_O"]))
         | ((r_S & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_S"]))
         | ((r_Z & 1) << (16 + data["X86"]["CondBitOffsets"]["G_CC_SHIFT_Z"]))
@@ -1341,7 +1356,6 @@ def x86g_calculate_aad_aam(state, flags_and_AX, opcode):
         | ((r_AH & 0xFF) << 8)
         | ((r_AL & 0xFF) << 0)
     )
-    return result
 
 
 #
@@ -1365,8 +1379,7 @@ def get_segdescr_limit(state, descriptor):
     limit = claripy.Concat(hi, lo).zero_extend(12)
     if (granularity == 0).is_true():
         return limit
-    else:
-        return (limit << 12) | 0xFFF
+    return (limit << 12) | 0xFFF
 
 
 def x86g_use_seg_selector(state, ldt, gdt, seg_selector, virtual_addr):
@@ -1544,9 +1557,7 @@ def armg_calculate_flag_n(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
     elif concrete_op == ARMG_CC_OP_SBB:
         res = cc_dep1 - cc_dep2 - (cc_dep3 ^ 1)
         flag = claripy.LShR(res, 31)
-    elif concrete_op == ARMG_CC_OP_LOGIC:
-        flag = claripy.LShR(cc_dep1, 31)
-    elif concrete_op == ARMG_CC_OP_MUL:
+    elif concrete_op in (ARMG_CC_OP_LOGIC, ARMG_CC_OP_MUL):
         flag = claripy.LShR(cc_dep1, 31)
     elif concrete_op == ARMG_CC_OP_MULL:
         flag = claripy.LShR(cc_dep2, 31)
@@ -1554,7 +1565,7 @@ def armg_calculate_flag_n(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
     if flag is not None:
         return flag
     l.error("Unknown cc_op %s (armg_calculate_flag_n)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def arm_zerobit(state, x):
@@ -1579,9 +1590,7 @@ def armg_calculate_flag_z(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
     elif concrete_op == ARMG_CC_OP_SBB:
         res = cc_dep1 - cc_dep2 - (cc_dep3 ^ 1)
         flag = arm_zerobit(state, res)
-    elif concrete_op == ARMG_CC_OP_LOGIC:
-        flag = arm_zerobit(state, cc_dep1)
-    elif concrete_op == ARMG_CC_OP_MUL:
+    elif concrete_op in (ARMG_CC_OP_LOGIC, ARMG_CC_OP_MUL):
         flag = arm_zerobit(state, cc_dep1)
     elif concrete_op == ARMG_CC_OP_MULL:
         flag = arm_zerobit(state, cc_dep1 | cc_dep2)
@@ -1590,7 +1599,7 @@ def armg_calculate_flag_z(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unknown cc_op %s (armg_calculate_flag_z)", concrete_op)
-    raise SimCCallError("Unknown cc_op %s" % concrete_op)
+    raise SimCCallError(f"Unknown cc_op {concrete_op}")
 
 
 def armg_calculate_flag_c(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -1617,16 +1626,14 @@ def armg_calculate_flag_c(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         )
     elif concrete_op == ARMG_CC_OP_LOGIC:
         flag = cc_dep2
-    elif concrete_op == ARMG_CC_OP_MUL:
-        flag = (claripy.LShR(cc_dep3, 1)) & 1
-    elif concrete_op == ARMG_CC_OP_MULL:
+    elif concrete_op in (ARMG_CC_OP_MUL, ARMG_CC_OP_MULL):
         flag = (claripy.LShR(cc_dep3, 1)) & 1
 
     if flag is not None:
         return flag
 
     l.error("Unknown cc_op %s (armg_calculate_flag_c)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def armg_calculate_flag_v(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -1653,16 +1660,14 @@ def armg_calculate_flag_v(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         flag = claripy.LShR(v, 31)
     elif concrete_op == ARMG_CC_OP_LOGIC:
         flag = cc_dep3
-    elif concrete_op == ARMG_CC_OP_MUL:
-        flag = cc_dep3 & 1
-    elif concrete_op == ARMG_CC_OP_MULL:
+    elif concrete_op in (ARMG_CC_OP_MUL, ARMG_CC_OP_MULL):
         flag = cc_dep3 & 1
 
     if flag is not None:
         return flag
 
     l.error("Unknown cc_op %s (armg_calculate_flag_v)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def armg_calculate_flags_nzcv(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -1728,7 +1733,7 @@ def armg_calculate_condition(state, cond_n_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unrecognized condition %d in armg_calculate_condition", concrete_cond)
-    raise SimCCallError("Unrecognized condition %d in armg_calculate_condition" % concrete_cond)
+    raise SimCCallError(f"Unrecognized condition {concrete_cond} in armg_calculate_condition")
 
 
 ARM64G_CC_SHIFT_N = 31
@@ -1811,7 +1816,7 @@ def arm64g_calculate_flag_n(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
             flag = flag.zero_extend(32)
         return flag
     l.error("Unknown cc_op %s (arm64g_calculate_flag_n)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def arm64_zerobit(state, x):
@@ -1865,7 +1870,7 @@ def arm64g_calculate_flag_z(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unknown cc_op %s (arm64g_calculate_flag_z)", concrete_op)
-    raise SimCCallError("Unknown cc_op %s" % concrete_op)
+    raise SimCCallError(f"Unknown cc_op {concrete_op}")
 
 
 def arm64g_calculate_flag_c(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -1899,7 +1904,7 @@ def arm64g_calculate_flag_c(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unknown cc_op %s (arm64g_calculate_flag_c)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def arm64g_calculate_flag_v(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -1957,7 +1962,7 @@ def arm64g_calculate_flag_v(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unknown cc_op %s (arm64g_calculate_flag_v)", cc_op)
-    raise SimCCallError("Unknown cc_op %s" % cc_op)
+    raise SimCCallError(f"Unknown cc_op {cc_op}")
 
 
 def arm64g_calculate_flags_nzcv(state, cc_op, cc_dep1, cc_dep2, cc_dep3):
@@ -2018,7 +2023,7 @@ def arm64g_calculate_condition(state, cond_n_op, cc_dep1, cc_dep2, cc_dep3):
         return flag
 
     l.error("Unrecognized condition %d in arm64g_calculate_condition", concrete_cond)
-    raise SimCCallError("Unrecognized condition %d in arm64g_calculate_condition" % concrete_cond)
+    raise SimCCallError(f"Unrecognized condition {concrete_cond} in arm64g_calculate_condition")
 
 
 #
@@ -2047,11 +2052,10 @@ def _get_flags(state) -> claripy.ast.bv.BV:
     except CCallMultivaluedException as e:
         cases, to_replace = e.args
         args = [cc_op, cc_dep1, cc_dep2, cc_ndep]
-        for i, arg in enumerate(args):
-            if arg is to_replace:
-                break
-        else:
-            raise errors.UnsupportedCCallError("Trying to concretize a value which is not an argument")
+        try:
+            i = args.index(to_replace)
+        except ValueError as ve:
+            raise errors.UnsupportedCCallError("Trying to concretize a value which is not an argument") from ve
         return claripy.ite_cases([(case, func(state, *args[:i], value_, *args[i + 1 :])) for case, value_ in cases], 0)
 
 
@@ -2073,8 +2077,7 @@ def _concat_flags(nbits, flags_vec):
     for offset, bit in flags_vec:
         current_position = nbits - 1 - result.length
         result = result.concat(claripy.BVV(0, current_position - offset), bit)
-    result = result.concat(claripy.BVV(0, nbits - result.length))
-    return result
+    return result.concat(claripy.BVV(0, nbits - result.length))
 
 
 def _get_nbits(cc_str):
@@ -2092,7 +2095,3 @@ def _get_nbits(cc_str):
     elif cc_str.endswith("64"):
         nbits = 64
     return nbits
-
-
-from angr.errors import SimError, SimCCallError
-from angr.sim_options import USE_SIMPLIFIED_CCALLS
