@@ -1,16 +1,23 @@
 # pylint:disable=raise-missing-from
-from typing import Dict, Generator, Optional, Set
+from __future__ import annotations
+
+import contextlib
+from collections.abc import Generator
 import logging
 import collections.abc
 import re
+import weakref
+import bisect
+import os
 from sortedcontainers import SortedDict
 
 import networkx
 
 from archinfo.arch_soot import SootMethodDescriptor
+import cle
 
-from ...errors import SimEngineError
-from ..plugin import KnowledgeBasePlugin
+from angr.errors import SimEngineError
+from angr.knowledge_plugins.plugin import KnowledgeBasePlugin
 from .function import Function
 from .soot_function import SootFunction
 
@@ -28,7 +35,7 @@ class FunctionDict(SortedDict):
     """
 
     def __init__(self, backref, *args, **kwargs):
-        self._backref = backref
+        self._backref = weakref.proxy(backref) if backref is not None else None
         self._key_types = kwargs.pop("key_types", int)
         super().__init__(*args, **kwargs)
 
@@ -36,18 +43,17 @@ class FunctionDict(SortedDict):
         try:
             return super().__getitem__(addr)
         except KeyError as ex:
-            if not isinstance(addr, self._key_types):
-                raise TypeError("FunctionDict only supports %s as key type" % self._key_types) from ex
+            if isinstance(addr, bool) or not isinstance(addr, self._key_types):
+                raise TypeError(f"FunctionDict only supports {self._key_types} as key type") from ex
 
             if isinstance(addr, SootMethodDescriptor):
                 t = SootFunction(self._backref, addr)
             else:
                 t = Function(self._backref, addr)
-            try:
+            with contextlib.suppress(Exception):
                 self[addr] = t
-            except Exception:  # pylint:disable=broad-except
-                pass
-            self._backref._function_added(t)
+            if self._backref is not None:
+                self._backref._function_added(t)
             return t
 
     def get(self, addr):
@@ -56,21 +62,21 @@ class FunctionDict(SortedDict):
     def floor_addr(self, addr):
         try:
             return next(self.irange(maximum=addr, reverse=True))
-        except StopIteration:
-            raise KeyError(addr)
+        except StopIteration as err:
+            raise KeyError(addr) from err
 
     def ceiling_addr(self, addr):
         try:
             return next(self.irange(minimum=addr, reverse=False))
-        except StopIteration:
-            raise KeyError(addr)
+        except StopIteration as err:
+            raise KeyError(addr) from err
 
     def __setstate__(self, state):
         for v, k in state.items():
             self[k] = v
 
     def __getstate__(self):
-        return {v: k for (v, k) in self.items()}
+        return dict(self.items())
 
 
 class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
@@ -83,14 +89,20 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         super().__init__(kb=kb)
         self.function_address_types = self._kb._project.arch.function_address_types
         self.address_types = self._kb._project.arch.address_types
-        self._function_map: Dict[int, Function] = FunctionDict(self, key_types=self.function_address_types)
-        self.function_addrs_set: Set = set()
+        self._function_map: FunctionDict[int, Function] = FunctionDict(self, key_types=self.function_address_types)
+        self.function_addrs_set: set = set()
         self.callgraph = networkx.MultiDiGraph()
         self.block_map = {}
 
         # Registers used for passing arguments around
         self._arg_registers = kb._project.arch.argument_registers
 
+        # local PLT dictionary cache
+        self._rplt_cache_ranges: None | list[tuple[int, int]] = None
+        self._rplt_cache: None | set[int] = None
+        # local binary name cache: min_addr -> (max_addr, binary_name)
+        self._binname_cache: None | SortedDict[int, tuple[int, str | None]] = None
+
     def __setstate__(self, state):
         self._kb = state["_kb"]
         self.function_address_types = state["function_address_types"]
@@ -99,12 +111,12 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         self.callgraph = state["callgraph"]
         self.block_map = state["block_map"]
 
-        self._function_map._backref = self
+        self._function_map._backref = weakref.proxy(self)
         for func in self._function_map.values():
             func._function_manager = self
 
     def __getstate__(self):
-        s = {
+        return {
             "_kb": self._kb,
             "function_address_types": self.function_address_types,
             "address_types": self.address_types,
@@ -112,7 +124,6 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
             "callgraph": self.callgraph,
             "block_map": self.block_map,
         }
-        return s
 
     def copy(self):
         fm = FunctionManager(self._kb)
@@ -130,17 +141,69 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         self.callgraph = networkx.MultiDiGraph()
         self.block_map.clear()
         self.function_addrs_set = set()
+        # cache
+        self._rplt_cache = None
+        self._rplt_cache_ranges = None
+        self._binname_cache = None
 
-    def _genenare_callmap_sif(self, filepath):
+    def _genenate_callmap_sif(self, filepath):
         """
         Generate a sif file from the call map.
 
         :param filepath:    Path of the sif file
         :return:            None
         """
-        with open(filepath, "wb") as f:
-            for src, dst in self.callgraph.edges():
-                f.write(f"{src:#x}\tDirectEdge\t{dst:#x}\n")
+        with open(filepath, "w", encoding="utf-8") as f:
+            f.writelines(f"{src:#x}\tDirectEdge\t{dst:#x}\n" for src, dst in self.callgraph.edges())
+
+    def _addr_in_plt_cached_ranges(self, addr: int) -> bool:
+        if self._rplt_cache_ranges is None:
+            return False
+        pos = bisect.bisect_left(self._rplt_cache_ranges, addr, key=lambda x: x[0])
+        return pos > 0 and self._rplt_cache_ranges[pos - 1][0] <= addr < self._rplt_cache_ranges[pos - 1][1]
+
+    def is_plt_cached(self, addr: int) -> bool:
+        # check if the addr is in the cache range
+        if not self._addr_in_plt_cached_ranges(addr):
+            # find the object containing this addr
+            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
+            if obj is None:
+                return False
+            if self._rplt_cache_ranges is None:
+                self._rplt_cache_ranges = []
+            obj_range = obj.min_addr, obj.max_addr
+            idx = bisect.bisect_left(self._rplt_cache_ranges, obj_range)
+            if not (idx < len(self._rplt_cache_ranges) and self._rplt_cache_ranges[idx] == obj_range):
+                self._rplt_cache_ranges.insert(idx, obj_range)
+            if isinstance(obj, cle.MetaELF):
+                if self._rplt_cache is None:
+                    self._rplt_cache = set()
+                self._rplt_cache |= set(obj.reverse_plt)
+
+        return addr in self._rplt_cache if self._rplt_cache is not None else False
+
+    def _binname_cache_get_addr_base(self, addr: int) -> int | None:
+        if self._binname_cache is None:
+            return None
+        try:
+            base_addr = next(self._binname_cache.irange(maximum=addr, reverse=True))
+        except StopIteration:
+            return None
+        return base_addr if base_addr <= addr < self._binname_cache[base_addr][0] else None
+
+    def get_binary_name_cached(self, addr: int) -> str | None:
+        base_addr = self._binname_cache_get_addr_base(addr)
+        if base_addr is None:
+            # not cached; cache it first
+            obj = self._kb._project.loader.find_object_containing(addr, membership_check=False)
+            if obj is None:
+                return None
+            if self._binname_cache is None:
+                self._binname_cache = SortedDict()
+            binary_basename = os.path.basename(obj.binary) if obj.binary else None
+            self._binname_cache[obj.min_addr] = obj.max_addr, binary_basename
+            base_addr = obj.min_addr
+        return self._binname_cache[base_addr][1] if self._binname_cache is not None else None
 
     def _add_node(self, function_addr, node, syscall=None, size=None):
         if isinstance(node, self.address_types):
@@ -148,7 +211,7 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         dst_func = self._function_map[function_addr]
         if syscall in (True, False):
             dst_func.is_syscall = syscall
-        dst_func._register_nodes(True, node)
+        dst_func._register_node(True, node)
         self.block_map[node.addr] = node
 
     def _add_call_to(
@@ -160,7 +223,7 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         syscall=None,
         stmt_idx=None,
         ins_addr=None,
-        return_to_outside=False,
+        return_to_outside: bool = False,
     ):
         """
         Add a call to a function.
@@ -172,7 +235,7 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         :param bool syscall:        If this is a call to a syscall or not.
         :param int stmt_idx:        ID of the statement where this call happens.
         :param int ins_addr:        Address of the instruction where this call happens.
-        :param bool return_to_outside:  True if the return of the call is considered going to outside of the current
+        :param return_to_outside:  True if the return of the call is considered going to outside of the current
                                         function.
         :return:                    None
         """
@@ -306,16 +369,16 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         try:
             _ = self[item]
             return True
-        except KeyError:
+        except (KeyError, TypeError):
             return False
 
     def __getitem__(self, k) -> Function:
         if isinstance(k, self.function_address_types):
             f = self.function(addr=k)
         elif type(k) is str:
-            f = self.function(name=k)
+            f = self.function(name=k) or self.function(name=k, check_previous_names=True)
         else:
-            raise ValueError("FunctionManager.__getitem__ does not support keys of type %s" % type(k))
+            raise ValueError(f"FunctionManager.__getitem__ does not support keys of type {type(k)}")
 
         if f is None:
             raise KeyError(k)
@@ -350,9 +413,9 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
     def get_by_addr(self, addr) -> Function:
         return self._function_map.get(addr)
 
-    def get_by_name(self, name: str) -> Generator[Function, None, None]:
+    def get_by_name(self, name: str, check_previous_names: bool = False) -> Generator[Function]:
         for f in self._function_map.values():
-            if f.name == name:
+            if f.name == name or (check_previous_names and name in f.previous_names):
                 yield f
 
     def _function_added(self, func: Function):
@@ -406,12 +469,12 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
 
         try:
             prev_addr = self._function_map.floor_addr(addr)
-            return self._function_map[prev_addr]
+            return self._function_map.get(prev_addr)
 
         except KeyError:
             return None
 
-    def query(self, query: str) -> Optional[Function]:
+    def query(self, query: str, check_previous_names: bool = False) -> Function | None:
         """
         Query for a function using selectors to disambiguate. Supported variations:
 
@@ -430,19 +493,21 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
                 addr = int(matches.group(2), 0)
                 try:
                     func = self._function_map.get(addr)
-                    if func.name == name:
+                    if func.name == name or (check_previous_names and name in func.previous_names):
                         return func
                 except KeyError:
                     pass
 
             obj_name = selector or self._kb._project.loader.main_object.binary_basename
-            for func in self.get_by_name(name):
+            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                 if func.binary_name == obj_name:
                     return func
 
         return None
 
-    def function(self, addr=None, name=None, create=False, syscall=False, plt=None) -> Optional[Function]:
+    def function(
+        self, addr=None, name=None, check_previous_names=False, create=False, syscall=False, plt=None
+    ) -> Function | None:
         """
         Get a function object from the function manager.
 
@@ -457,6 +522,19 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
         :return: The Function instance, or None if the function is not found and create is False.
         :rtype: Function or None
         """
+        if name is not None and name.startswith("sub_"):
+            # first check if a function with the specified name exists
+            for func in self.get_by_name(name, check_previous_names=check_previous_names):
+                if plt is None or func.is_plt == plt:
+                    return func
+
+            # then enter the syntactic sugar mode
+            try:
+                addr = int(name.split("_")[-1], 16)
+                name = None
+            except ValueError:
+                pass
+
         if addr is not None:
             try:
                 f = self._function_map.get(addr)
@@ -472,11 +550,11 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
                         f.is_syscall = True
                     return f
         elif name is not None:
-            func = self.query(name)
+            func = self.query(name, check_previous_names=check_previous_names)
             if func is not None:
                 return func
 
-            for func in self.get_by_name(name):
+            for func in self.get_by_name(name, check_previous_names=check_previous_names):
                 if plt is None or func.is_plt == plt:
                     return func
 
@@ -489,13 +567,22 @@ class FunctionManager(KnowledgeBasePlugin, collections.abc.Mapping):
 
     def rebuild_callgraph(self):
         self.callgraph = networkx.MultiDiGraph()
-        for func_addr in self._function_map.keys():
+        cfg = self._kb.cfgs.get_most_accurate()
+        for func_addr in self._function_map:
             self.callgraph.add_node(func_addr)
         for func in self._function_map.values():
             if func.block_addrs_set:
                 for node in func.transition_graph.nodes():
                     if isinstance(node, Function):
                         self.callgraph.add_edge(func.addr, node.addr)
+                    else:
+                        cfgnode = cfg.get_any_node(node.addr)
+                        if (
+                            cfgnode is not None
+                            and cfgnode.function_address is not None
+                            and cfgnode.function_address != func.addr
+                        ):
+                            self.callgraph.add_edge(func.addr, cfgnode.function_address)
 
 
 KnowledgeBasePlugin.register_default("functions", FunctionManager)
