@@ -1,3 +1,4 @@
+from __future__ import annotations
 import logging
 import os
 import types
@@ -6,20 +7,19 @@ import pickle
 import string
 from collections import defaultdict
 from pathlib import Path
-from typing import Dict, Any, Optional
+from typing import Any, cast
 
 import archinfo
 from archinfo.arch_soot import SootAddressDescriptor, ArchSoot
 import cle
 from .sim_procedure import SimProcedure
 
-from .misc.ux import deprecated
 from .errors import AngrNoPluginError
 
 l = logging.getLogger(name=__name__)
 
 
-def load_shellcode(shellcode, arch, start_offset=0, load_address=0, thumb=False, **kwargs):
+def load_shellcode(shellcode: bytes | str, arch, start_offset=0, load_address=0, thumb=False, **kwargs):
     """
     Load a new project based on a snippet of assembly or bytecode.
 
@@ -31,13 +31,18 @@ def load_shellcode(shellcode, arch, start_offset=0, load_address=0, thumb=False,
     """
     if not isinstance(arch, archinfo.Arch):
         arch = archinfo.arch_from_id(arch)
-    if type(shellcode) is str:
-        shellcode = arch.asm(shellcode, load_address, thumb=thumb)
+    if isinstance(shellcode, str):
+        shellcode_bytes = arch.asm(shellcode, load_address, thumb=thumb)
+        if shellcode_bytes is None:
+            raise ValueError("Could not assemble shellcode")
+        assert isinstance(shellcode_bytes, bytes)
+    else:
+        shellcode_bytes = shellcode
     if thumb:
         start_offset |= 1
 
     return Project(
-        BytesIO(shellcode),
+        BytesIO(shellcode_bytes),
         main_opts={
             "backend": "blob",
             "arch": arch,
@@ -92,7 +97,6 @@ class Project:
     :type storage:      defaultdict(list)
     """
 
-    analyses: "AnalysesHub"
     arch: archinfo.Arch
 
     def __init__(
@@ -106,10 +110,10 @@ class Project:
         arch=None,
         simos=None,
         engine=None,
-        load_options: Dict[str, Any] = None,
+        load_options: dict[str, Any] | None = None,
         translation_cache=True,
         selfmodifying_code: bool = False,
-        support_selfmodifying_code: Optional[bool] = None,  # deprecated. use selfmodifying_code instead
+        support_selfmodifying_code: bool | None = None,  # deprecated. use selfmodifying_code instead
         store_function=None,
         load_function=None,
         analyses_preset=None,
@@ -138,7 +142,7 @@ class Project:
             self.filename = None
             self.loader = cle.Loader(thing, **load_options)
         elif not isinstance(thing, (str, Path)) or not os.path.exists(thing) or not os.path.isfile(thing):
-            raise Exception("Not a valid binary file: %s" % repr(thing))
+            raise Exception(f"Not a valid binary file: {thing!r}")
         else:
             # use angr's loader, provided by cle
             l.info("Loading binary %s", thing)
@@ -172,14 +176,14 @@ class Project:
 
         # It doesn't make any sense to have auto_load_libs
         # if you have the concrete target, let's warn the user about this.
-        if self.concrete_target and load_options.get("auto_load_libs", None):
+        if self.concrete_target and load_options.get("auto_load_libs"):
             l.critical(
                 "Incompatible options selected for this project, please disable auto_load_libs if "
                 "you want to use a concrete target."
             )
             raise Exception("Incompatible options for the project")
 
-        if self.concrete_target and self.arch.name not in ["X86", "AMD64", "ARMHF", "MIPS32"]:
+        if self.concrete_target and self.arch.name not in ["X86", "AMD64", "ARMHF", "ARMEL", "MIPS32"]:
             l.critical("Concrete execution does not support yet the selected architecture. Aborting.")
             raise Exception("Incompatible options for the project")
 
@@ -201,29 +205,20 @@ class Project:
             )
             self.selfmodifying_code = bool(support_selfmodifying_code)
 
-        if self.selfmodifying_code:
-            if self._translation_cache is True:
-                self._translation_cache = False
-                l.warning("Disabling IRSB translation cache because support for self-modifying code is enabled.")
+        if self.selfmodifying_code and self._translation_cache is True:
+            self._translation_cache = False
+            l.warning("Disabling IRSB translation cache because support for self-modifying code is enabled.")
 
         self.entry = self.loader.main_object.entry
         self.storage = defaultdict(list)
         self.store_function = store_function or self._store
         self.load_function = load_function or self._load
 
-        # Step 4: Set up the project's hubs
-        # Step 4.1 Factory
-        self.factory = AngrObjectFactory(self, default_engine=engine)
+        # FIXME: BIG HACK
+        if isinstance(self.filename, str) and self.filename.endswith(".xcal") and not simos:
+            simos = "snimmuc_nxp"
 
-        # Step 4.2: Analyses
-        self._analyses_preset = analyses_preset
-        self.analyses = None
-        self._initialize_analyses_hub()
-
-        # Step 4.3: ...etc
-        self.kb = KnowledgeBase(self, name="global")
-
-        # Step 5: determine the guest OS
+        # Step 4: determine the guest OS
         if isinstance(simos, type) and issubclass(simos, SimOS):
             self.simos = simos(self)  # pylint:disable=invalid-name
         elif isinstance(simos, str):
@@ -233,11 +228,25 @@ class Project:
         else:
             raise ValueError("Invalid OS specification or non-matching architecture.")
 
+        # Step 5: Set up the project's hubs
+        # Step 5.1 Factory
+        self.factory = AngrObjectFactory(self, default_engine=engine)
+
+        # Step 5.2: Analyses
+        self._analyses_preset = analyses_preset
+        self._analyses = None
+        self._initialize_analyses_hub()
+
+        # Step 5.3: ...etc
+        self._knowledge_bases = {"default": KnowledgeBase(self, name="global")}
+
         self.is_java_project = isinstance(self.arch, ArchSoot)
-        self.is_java_jni_project = isinstance(self.arch, ArchSoot) and self.simos.is_javavm_with_jni_support
+        self.is_java_jni_project = isinstance(self.arch, ArchSoot) and getattr(
+            self.simos, "is_javavm_with_jni_support", False
+        )
 
         # Step 6: Register simprocedures as appropriate for library functions
-        if isinstance(self.arch, ArchSoot) and self.simos.is_javavm_with_jni_support:
+        if isinstance(self.arch, ArchSoot) and getattr(self.simos, "is_javavm_with_jni_support", False):
             # If we execute a Java archive that includes native JNI libraries,
             # we need to use the arch of the native simos for all (native) sim
             # procedures.
@@ -250,12 +259,39 @@ class Project:
         # Step 7: Run OS-specific configuration
         self.simos.configure_project()
 
+    @property
+    def kb(self):
+        return self._knowledge_bases["default"]
+
+    @kb.setter
+    def kb(self, kb):
+        self._knowledge_bases["default"] = kb
+
+    def get_kb(self, name):
+        try:
+            return self._knowledge_bases[name]
+        except KeyError:
+            kb = KnowledgeBase(self, name)
+            self._knowledge_bases[name] = kb
+            return kb
+
+    @property
+    def analyses(self) -> AnalysesHubWithDefault:
+        result = self._analyses
+        if result is None:
+            raise ValueError("Cannot access analyses this early in project lifecycle")
+        return result
+
+    @analyses.setter
+    def analyses(self, v: AnalysesHubWithDefault):
+        self._analyses = v
+
     def _initialize_analyses_hub(self):
         """
         Initializes self.analyses using a given preset.
         """
-        self.analyses = AnalysesHub(self)
-        self.analyses.use_plugin_preset(self._analyses_preset if self._analyses_preset is not None else "default")
+        self._analyses = cast("AnalysesHubWithDefault", AnalysesHub(self))
+        self._analyses.use_plugin_preset(self._analyses_preset if self._analyses_preset is not None else "default")
 
     def _register_object(self, obj, sim_proc_arch):
         """
@@ -264,18 +300,26 @@ class Project:
 
         # Step 1: get the set of libraries we are allowed to use to resolve unresolved symbols
         missing_libs = []
+        missing_wincore_dlls = False
         for lib_name in self.loader.missing_dependencies:
             try:
-                missing_libs.append(SIM_LIBRARIES[lib_name])
+                missing_libs.extend(SIM_LIBRARIES[lib_name])
             except KeyError:
                 l.info("There are no simprocedures for missing library %s :(", lib_name)
+                if lib_name.startswith("api-ms-win-"):
+                    missing_wincore_dlls = True
+        if missing_wincore_dlls and "kernel32.dll" not in self.loader.missing_dependencies:
+            # some of the missing api-ms-win-*.dll libraries are actually provided by kernel32.dll
+            missing_libs.extend(SIM_LIBRARIES["kernel32.dll"])
+
         # additionally provide libraries we _have_ loaded as a fallback fallback
         # this helps in the case that e.g. CLE picked up a linux arm libc to satisfy an android arm binary
         for lib in self.loader.all_objects:
-            if lib.provides in SIM_LIBRARIES:
-                simlib = SIM_LIBRARIES[lib.provides]
-                if simlib not in missing_libs:
-                    missing_libs.append(simlib)
+            if lib.provides is not None and lib.provides in SIM_LIBRARIES:
+                simlibs = SIM_LIBRARIES[lib.provides]
+                for simlib in simlibs:
+                    if simlib not in missing_libs:
+                        missing_libs.append(simlib)
 
         # Step 2: Categorize every "import" symbol in each object.
         # If it's IGNORED, mark it for stubbing
@@ -328,11 +372,13 @@ class Project:
                     owner_name = owner_name.lower()
                 if owner_name not in SIM_LIBRARIES:
                     continue
-                sim_lib = SIM_LIBRARIES[owner_name]
-                if not sim_lib.has_implementation(export.name):
-                    continue
-                l.info("Using builtin SimProcedure for %s from %s", export.name, sim_lib.name)
-                self.hook_symbol(export.rebased_addr, sim_lib.get(export.name, sim_proc_arch))
+                sim_libs = SIM_LIBRARIES[owner_name]
+                for sim_lib in sim_libs:
+                    if not sim_lib.has_implementation(export.name):
+                        continue
+                    l.info("Using builtin SimProcedure for %s from %s", export.name, sim_lib.name)
+                    self.hook_symbol(export.rebased_addr, sim_lib.get(export.name, sim_proc_arch))
+                    break
 
             # Step 2.3: If 2.2 didn't work, check if the symbol wants to be resolved
             # by a library we already know something about. Resolve it appropriately.
@@ -341,7 +387,7 @@ class Project:
             # we still want to try as hard as we can to figure out where it comes from
             # so we can get the calling convention as close to right as possible.
             elif reloc.resolvewith is not None and reloc.resolvewith in SIM_LIBRARIES:
-                sim_lib = SIM_LIBRARIES[reloc.resolvewith]
+                sim_lib = sorted(SIM_LIBRARIES[reloc.resolvewith], key=lambda lib: lib.has_prototype(export.name))[-1]
                 if self._check_user_blacklists(export.name):
                     if not func.is_weak:
                         l.info("Using stub SimProcedure for unresolved %s from %s", func.name, sim_lib.name)
@@ -373,7 +419,7 @@ class Project:
                         if export.name and export.name.startswith("_Z"):
                             # GNU C++ name. Use a C++ library to create the stub
                             if "libstdc++.so" in SIM_LIBRARIES:
-                                the_lib = SIM_LIBRARIES["libstdc++.so"]
+                                the_lib = SIM_LIBRARIES["libstdc++.so"][0]
                             else:
                                 l.critical(
                                     "Does not find any C++ library in SIM_LIBRARIES. We may not correctly "
@@ -403,20 +449,17 @@ class Project:
         """
         # First, filter the SIM_LIBRARIES to a reasonable subset based on the hint
         if hint == "win":
-            hinted_libs = filter(lambda lib: lib if lib.endswith(".dll") else None, SIM_LIBRARIES)
+            hinted_libs = [lib for lib in SIM_LIBRARIES if lib.endswith(".dll")]
         else:
-            hinted_libs = filter(lambda lib: lib if ".so" in lib else None, SIM_LIBRARIES)
+            hinted_libs = [lib for lib in SIM_LIBRARIES if ".so" in lib]
 
         for lib in hinted_libs:
-            if SIM_LIBRARIES[lib].has_implementation(f.name):
-                l.debug("Found implementation for %s in %s", f, lib)
-                if f.resolvedby:
-                    hook_at = f.resolvedby.rebased_addr
-                else:
-                    # ????
-                    hook_at = f.relative_addr
-                self.hook_symbol(hook_at, (SIM_LIBRARIES[lib].get(f.name, self.arch)))
-                return True
+            for simlib in SIM_LIBRARIES[lib]:
+                if simlib.has_implementation(f.name):
+                    l.debug("Found implementation for %s in %s", f, lib)
+                    hook_at = f.resolvedby.rebased_addr if f.resolvedby else f.relative_addr  # ????
+                    self.hook_symbol(hook_at, (simlib.get(f.name, self.arch)))
+                    return True
 
         l.debug("Could not find matching SimProcedure for %s, ignoring.", f.name)
         return False
@@ -435,7 +478,7 @@ class Project:
 
     @staticmethod
     def _addr_to_str(addr):
-        return "%s" % repr(addr) if isinstance(addr, SootAddressDescriptor) else "%#x" % addr
+        return f"{addr!r}" if isinstance(addr, SootAddressDescriptor) else f"{addr:#x}"
 
     #
     # Public methods
@@ -443,7 +486,7 @@ class Project:
     #
 
     # pylint: disable=inconsistent-return-statements
-    def hook(self, addr, hook=None, length=0, kwargs=None, replace=False):
+    def hook(self, addr, hook=None, length=0, kwargs=None, replace: bool | None = False):
         """
         Hook a section of code with a custom function. This is used internally to provide symbolic
         summaries of library functions, and can be used to instrument execution or to modify
@@ -473,7 +516,7 @@ class Project:
         """
         if hook is None:
             # if we haven't been passed a thing to hook with, assume we're being used as a decorator
-            return self._hook_decorator(addr, length=length, kwargs=kwargs)
+            return self._hook_decorator(addr, length=length, kwargs=kwargs, replace=replace)
 
         if kwargs is None:
             kwargs = {}
@@ -487,17 +530,26 @@ class Project:
                 l.warning(
                     "Address is already hooked, during hook(%s, %s). Not re-hooking.", self._addr_to_str(addr), hook
                 )
-                return
+                return None
             else:
                 l.warning("Address is already hooked, during hook(%s, %s). Re-hooking.", self._addr_to_str(addr), hook)
 
         if isinstance(hook, type):
-            raise TypeError("Please instanciate your SimProcedure before hooking with it")
+            raise TypeError("Please instantiate your SimProcedure before hooking with it")
 
         if callable(hook):
             hook = SIM_PROCEDURES["stubs"]["UserHook"](user_func=hook, length=length, **kwargs)
 
+        # if not hook.is_stub and hook.guessed_prototype:
+        #    l.error(
+        #        "You need to instantiate %s with a prototype. "
+        #        "It will attempt to guess one but the consequences of guessing wrong are dire.",
+        #        hook,
+        #    )
+        #    l.error("Consider also using angr.SIM_LIBRARIES instead of angr.SIM_PROCEDURES or angr.procedures.")
+
         self._sim_procedures[addr] = hook
+        return None
 
     def is_hooked(self, addr) -> bool:
         """
@@ -508,7 +560,7 @@ class Project:
         """
         return addr in self._sim_procedures
 
-    def hooked_by(self, addr) -> Optional[SimProcedure]:
+    def hooked_by(self, addr) -> SimProcedure | None:
         """
         Returns the current hook for `addr`.
 
@@ -535,7 +587,7 @@ class Project:
 
         del self._sim_procedures[addr]
 
-    def hook_symbol(self, symbol_name, simproc, kwargs=None, replace=None):
+    def hook_symbol(self, symbol_name, simproc, kwargs=None, replace: bool | None = None):
         """
         Resolve a dependency in a binary. Looks up the address of the given symbol, and then hooks that
         address. If the symbol was not available in the loaded libraries, this address may be provided
@@ -563,12 +615,13 @@ class Project:
                 # it could be a previously unresolved weak symbol..?
                 new_sym = None
                 for reloc in self.loader.find_relevant_relocations(symbol_name):
+                    assert reloc.symbol is not None
                     if not reloc.symbol.is_weak:
                         raise Exception("Symbol is strong but we couldn't find its resolution? Report to @rhelmot.")
                     if new_sym is None:
                         new_sym = self.loader.extern_object.make_extern(symbol_name)
                     reloc.resolve(new_sym)
-                    reloc.relocate([])
+                    reloc.relocate()
 
                 if new_sym is None:
                     l.error("Could not find symbol %s", symbol_name)
@@ -585,7 +638,7 @@ class Project:
         self.hook(hook_addr, simproc, kwargs=kwargs, replace=replace)
         return hook_addr
 
-    def symbol_hooked_by(self, symbol_name) -> Optional[SimProcedure]:
+    def symbol_hooked_by(self, symbol_name) -> SimProcedure | None:
         """
         Return the SimProcedure, if it exists, for the given symbol name.
 
@@ -596,8 +649,10 @@ class Project:
         sym = self.loader.find_symbol(symbol_name)
         if sym is None:
             l.warning("Could not find symbol %s", symbol_name)
-            return False
+            return None
         hook_addr, _ = self.simos.prepare_function_symbol(symbol_name, basic_addr=sym.rebased_addr)
+        if not self.is_hooked(hook_addr):
+            return None
         return self.hooked_by(hook_addr)
 
     def is_symbol_hooked(self, symbol_name):
@@ -680,10 +735,7 @@ class Project:
         simulation manager.
         """
 
-        if args:
-            state = args[0]
-        else:
-            state = self.factory.full_init_state(**kwargs)
+        state = args[0] if args else self.factory.full_init_state(**kwargs)
 
         pg = self.factory.simulation_manager(state)
         self._executing = True
@@ -699,7 +751,7 @@ class Project:
     # Private methods related to hooking
     #
 
-    def _hook_decorator(self, addr, length=0, kwargs=None):
+    def _hook_decorator(self, addr, length=0, kwargs=None, replace=False):
         """
         Return a function decorator that allows easy hooking. Please refer to hook() for its usage.
 
@@ -707,7 +759,7 @@ class Project:
         """
 
         def hook_decorator(func):
-            self.hook(addr, func, length=length, kwargs=kwargs)
+            self.hook(addr, func, length=length, kwargs=kwargs, replace=replace)
             return func
 
         return hook_decorator
@@ -717,12 +769,12 @@ class Project:
     #
 
     def __getstate__(self):
+        store_func, load_func = self.store_function, self.load_function
         try:
-            store_func, load_func = self.store_function, self.load_function
             self.store_function, self.load_function = None, None
             # ignore analyses. we re-initialize analyses when restoring from pickling so that we do not lose any newly
             # added analyses classes
-            d = {
+            return {
                 k: v
                 for k, v in self.__dict__.items()
                 if k
@@ -730,7 +782,6 @@ class Project:
                     "analyses",
                 }
             }
-            return d
         finally:
             self.store_function, self.load_function = store_func, load_func
 
@@ -773,10 +824,8 @@ class Project:
             if all(c in string.printable for c in container) and os.path.exists(container):
                 with open(container, "rb") as f:
                     return pickle.load(f)
-
-            # If container is a pickle string.
             else:
-                return pickle.loads(container)
+                raise ValueError(container)
 
         # If container is an open file
         elif isinstance(container, IOBase):
@@ -790,18 +839,9 @@ class Project:
     def __repr__(self):
         return "<Project %s>" % (self.filename if self.filename is not None else "loaded from stream")
 
-    #
-    # Compatibility
-    #
-
-    @property
-    @deprecated(replacement="simos")
-    def _simos(self):
-        return self.simos
-
 
 from .factory import AngrObjectFactory
-from angr.simos import SimOS, os_mapping
-from .analyses.analysis import AnalysesHub
+from .simos import SimOS, os_mapping
+from .analyses.analysis import AnalysesHub, AnalysesHubWithDefault
 from .knowledge_base import KnowledgeBase
 from .procedures import SIM_PROCEDURES, SIM_LIBRARIES
